from ..utils import serialize_num


class Width(object):
    def __init__(self, width):
        self.width = width

    def serialize(self, output):
        output.write(b"<")
        output.write(serialize_num(self.width))
        output.write(b">")



from py_hcl.transformer.pyhcl_to_firrtl.context import Context
from py_hcl.transformer.pyhcl_to_firrtl.conv_port import convert_ports
from py_hcl.transformer.pyhcl_to_firrtl.conv_stmt import convert_stmt
from py_hcl.transformer.pyhcl_to_firrtl.utils import build_reserve_name, \
    build_io_name, get_io_obj
from py_hcl.core.expr.io import IO
from py_hcl.core.module.packed_module import PackedModule
from py_hcl.core.module_factory.inherit_list.named_expr import NamedExprChain
from py_hcl.core.module_factory.inherit_list.stmt_holder import StmtChain
from py_hcl.firrtl_ir.stmt.block import Block
from py_hcl.firrtl_ir.stmt.defn.module import DefModule


def convert_module(packed_module: PackedModule):
    Context.expr_id_to_name.update(
        flatten_named_expr_chain(packed_module.named_expr_chain))

    name = packed_module.name
    raw_ports = flatten_io_chain(get_io_obj(packed_module))
    ports = convert_ports(raw_ports)

    stmts = flatten_statement_chain(packed_module.statement_chain)
    final_stmts = [ss for s in stmts for ss in convert_stmt(s)]

    module = DefModule(name, ports, Block(final_stmts))
    Context.modules[name] = module
    return module


def flatten_named_expr_chain(named_expr_chain: NamedExprChain):
    named_exprs = {}
    node = named_expr_chain.named_expr_chain_head

    while True:
        holder = node.named_expr_holder
        for k, v in holder.named_expression_table.items():
            named_exprs[k] = build_reserve_name(holder.module_name, v)

        if not hasattr(node, "next_node"):
            break
        node = node.next_node

    return named_exprs


def flatten_statement_chain(statement_chain: StmtChain):
    stmts = []
    node = statement_chain.stmt_chain_head

    while True:
        holder = node.stmt_holder
        for stmt in reversed(holder.top_statement.statements):
            stmts.append(stmt)

        if not hasattr(node, "next_node"):
            break
        node = node.next_node

    return list(reversed(stmts))


def flatten_io_chain(io: IO):
    ports = {}
    node = io.io_chain_head

    while True:
        holder = node.io_holder
        # reverse the dict order
        for k in list(holder.named_ports.keys())[::-1]:
            v = holder.named_ports[k]
            ports[build_io_name(holder.module_name, k)] = v

        if not hasattr(node, "next_node"):
            break
        node = node.next_node

    return {k: ports[k] for k in list(ports.keys())[::-1]}



from py_hcl.firrtl_ir.expr.prim_ops import Xor
from py_hcl.firrtl_ir.shortcuts import uw, u, w, s
from py_hcl.firrtl_ir.type import UIntType, SIntType
from tests.test_firrtl_ir.utils import serialize_equal
from .helper import OpCase, basis_tester, \
    encounter_error_tester, type_wrong_cases_2_args_gen, max_width


def args(*arg_types):
    class C:
        @staticmethod
        def tpe(res_type):
            return OpCase(Xor).arg_types(*arg_types).res_type(res_type)

    return C


xor_basis_cases = [
    args(UIntType, UIntType).tpe(lambda x, y: uw(max_width(x, y))),
    args(SIntType, SIntType).tpe(lambda x, y: uw(max_width(x, y))),
]

xor_type_wrong_cases = type_wrong_cases_2_args_gen(Xor)

xor_width_wrong_cases = [
    args(UIntType, UIntType).tpe(lambda x, y: uw(max_width(x, y) + 1)),
    args(SIntType, SIntType).tpe(lambda x, y: uw(max_width(x, y) + 1)),
    args(UIntType, UIntType).tpe(lambda x, y: uw(max_width(x, y) - 1)),
    args(SIntType, SIntType).tpe(lambda x, y: uw(max_width(x, y) - 1)),
]


def test_xor():
    basis_tester(xor_basis_cases)
    encounter_error_tester(xor_type_wrong_cases)
    encounter_error_tester(xor_width_wrong_cases)
    serialize_equal(Xor([u(20, w(5)), u(15, w(4))], uw(5)),
                    'xor(UInt<5>("h14"), UInt<4>("hf"))')
    serialize_equal(Xor([s(-20, w(6)), s(-15, w(5))], uw(6)),
                    'xor(SInt<6>("h-14"), SInt<5>("h-f"))')



def install_ops():
    import py_hcl.core.expr.add  # noqa: F401
    import py_hcl.core.expr.and_  # noqa: F401
    import py_hcl.core.expr.xor  # noqa: F401
    import py_hcl.core.expr.or_  # noqa: F401
    import py_hcl.core.stmt.connect  # noqa: F401
    import py_hcl.core.expr.field  # noqa: F401
    import py_hcl.core.expr.slice  # noqa: F401
    import py_hcl.core.expr.index  # noqa: F401
    import py_hcl.core.expr.convert  # noqa: F401
    import py_hcl.core.expr.extend  # noqa: F401



from ...utils import serialize_str
from .. import Statement


class InputPort(object):
    def __init__(self, name, tpe):
        self.name = name
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"input ")
        output.write(serialize_str(self.name))
        output.write(b" : ")
        self.tpe.serialize(output)


class OutputPort(object):
    def __init__(self, name, tpe):
        self.name = name
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"output ")
        output.write(serialize_str(self.name))
        output.write(b" : ")
        self.tpe.serialize(output)


class DefModule(Statement):
    def __init__(self, name, ports, body):
        self.name = name
        self.ports = ports
        self.body = body

    def serialize_stmt(self, output, indent):
        output.write(b"module ")
        output.write(serialize_str(self.name))
        output.write(b" :\n")
        indent += 1
        for port in self.ports:
            output.write(serialize_str("  " * indent))
            port.serialize(output)
            output.write(b"\n")
        output.write(b"\n")
        output.write(serialize_str("  " * indent))
        self.body.serialize_stmt(output, indent)


class DefExtModule(Statement):
    def __init__(self, name, ports, def_name):
        self.name = name
        self.ports = ports
        self.def_name = def_name

    def serialize_stmt(self, output, indent):
        output.write(b"extmodule ")
        output.write(serialize_str(self.name))
        output.write(b" :\n")
        indent += 1
        for port in self.ports:
            output.write(serialize_str("  " * indent))
            port.serialize(output)
            output.write(b"\n")
        output.write(b"\n")
        output.write(serialize_str("  " * indent))
        output.write(b"defname = ")
        output.write(serialize_str(self.def_name))



from ..utils import serialize_num


class UnknownType(object):
    def serialize(self, output):
        output.write(b"?")


class GroundType(object):
    pass


class UIntType(GroundType):
    def __init__(self, width):
        self.width = width

    def serialize(self, output):
        output.write(b"UInt")
        self.width.serialize(output)


class SIntType(GroundType):
    def __init__(self, width):
        self.width = width

    def serialize(self, output):
        output.write(b"SInt")
        self.width.serialize(output)


class ClockType(GroundType):
    def serialize(self, output):
        output.write(b"Clock")


class AggregateType(object):
    pass


class BundleType(AggregateType):
    def __init__(self, fields):
        self.fields = fields

    def serialize(self, output):
        output.write(b"{")
        for f in self.fields[:-1]:
            f.serialize(output)
            output.write(b", ")
        self.fields[-1].serialize(output)
        output.write(b"}")


class VectorType(AggregateType):
    def __init__(self, elem_type, size):
        self.elem_type = elem_type
        self.size = size

    def serialize(self, output):
        self.elem_type.serialize(output)
        output.write(b"[")
        output.write(serialize_num(self.size))
        output.write(b"]")



from py_hcl.core.expr import ExprHolder
from py_hcl.core.expr.error import ExprError
from py_hcl.core.expr.utils import assert_right_side
from py_hcl.core.hcl_ops import op_register
from py_hcl.core.stmt.connect import ConnSide
from py_hcl.core.type import HclType
from py_hcl.core.type.sint import SIntT
from py_hcl.core.type.uint import UIntT
from py_hcl.utils import json_serialize

to_bool = op_register('to_bool')
to_uint = op_register('to_uint')
to_sint = op_register('to_sint')


@json_serialize
class ToSInt(object):
    def __init__(self, expr):
        self.operation = "to_sint"
        self.ref_expr_id = expr.id


@json_serialize
class ToUInt(object):
    def __init__(self, expr):
        self.operation = "to_uint"
        self.ref_expr_id = expr.id


@to_bool(UIntT)
@assert_right_side
def _(uint):
    return uint[0]


@to_bool(SIntT)
@assert_right_side
def _(sint):
    return sint[0]


@to_bool(HclType)
def _(_0, *_):
    raise ExprError.op_type_err('to_bool', _0)


@to_uint(UIntT)
@assert_right_side
def _(uint):
    return uint


@to_uint(SIntT)
@assert_right_side
def _(sint):
    t = UIntT(sint.hcl_type.width)
    return ExprHolder(t, ConnSide.RT, ToUInt(sint))


@to_uint(HclType)
def _(_0, *_):
    raise ExprError.op_type_err('to_uint', _0)


@to_sint(UIntT)
@assert_right_side
def _(uint):
    t = SIntT(uint.hcl_type.width)
    return ExprHolder(t, ConnSide.RT, ToSInt(uint))


@to_sint(SIntT)
@assert_right_side
def _(sint):
    return sint


@to_sint(HclType)
def _(_0, *_):
    raise ExprError.op_type_err('to_sint', _0)



import pytest

from py_hcl.core.stmt import ConditionStatement
from py_hcl.core.stmt.connect import Connect
from py_hcl.core.stmt.error import StatementError
from py_hcl.core.stmt_factory.scope import ScopeType, ScopeManager
from py_hcl.dsl.branch import when, else_when, otherwise
from py_hcl.dsl.expr.io import IO
from py_hcl.dsl.expr.wire import Wire
from py_hcl.dsl.module import Module
from py_hcl.dsl.tpe.uint import U


def test_branch():
    class A(Module):
        io = IO()
        a = Wire(U.w(8))
        b = Wire(U.w(8))
        c = Wire(U.w(8))

        a <<= b
        with when(U(0)):
            a <<= b + c
            c <<= a
        with else_when(U(1)):
            b <<= a + c
            with when(U(0)):
                b <<= a
            with otherwise():
                c <<= a
        with otherwise():
            c <<= a + b
            c <<= a + b
            c <<= a + b
            c <<= a + b

    s = A.packed_module.statement_chain \
        .stmt_chain_head.stmt_holder.top_statement.statements
    assert len(s) == 2

    si = ScopeManager.get_scope_info(s[0].scope_id)
    assert si.scope_type == ScopeType.GROUND
    assert isinstance(s[0].statement, Connect)

    assert isinstance(s[1], ConditionStatement)
    assert len(s[1].seq_stmts) == 2

    assert len(s[1].alt_stmts) == 1
    assert isinstance(s[1].alt_stmts[0], ConditionStatement)
    assert len(s[1].alt_stmts[0].seq_stmts) == 2
    assert len(s[1].alt_stmts[0].alt_stmts) == 4

    assert isinstance(s[1].alt_stmts[0].seq_stmts[1], ConditionStatement)
    assert len(s[1].alt_stmts[0].seq_stmts[1].seq_stmts) == 1
    assert len(s[1].alt_stmts[0].seq_stmts[1].alt_stmts) == 1


def test_branch_syntax_error1():
    with pytest.raises(StatementError):
        class A(Module):
            io = IO()
            a = Wire(U.w(8))
            b = Wire(U.w(8))
            c = Wire(U.w(8))

            a <<= b
            with else_when(U(0)):
                b <<= a + c
            with otherwise():
                c <<= a + b


def test_branch_syntax_error2():
    with pytest.raises(StatementError):
        class A(Module):
            io = IO()
            a = Wire(U.w(8))
            b = Wire(U.w(8))
            c = Wire(U.w(8))

            with otherwise():
                c <<= a + b


def test_branch_syntax_error3():
    with pytest.raises(StatementError):
        class A(Module):
            io = IO()
            a = Wire(U.w(8))
            b = Wire(U.w(8))
            c = Wire(U.w(8))

            with when(U(0)):
                b <<= a + c
                with otherwise():
                    c <<= a + b


def test_branch_syntax_error4():
    with pytest.raises(StatementError):
        class A(Module):
            io = IO()
            a = Wire(U.w(8))
            b = Wire(U.w(8))
            c = Wire(U.w(8))

            with when(U(0)):
                b <<= a + c
                with else_when(U(1)):
                    c <<= a + b
            with otherwise():
                c <<= a + b


def test_branch_syntax_error5():
    with pytest.raises(StatementError):
        class A(Module):
            io = IO()
            a = Wire(U.w(8))
            b = Wire(U.w(8))
            c = Wire(U.w(8))

            with when(U(0)):
                b <<= a + c
            c <<= a + b
            with otherwise():
                c <<= a + b



import logging
from enum import Enum

from py_hcl.core.hcl_ops import op_register, op_apply
from py_hcl.core.stmt.error import StatementError
from py_hcl.core.stmt_factory.trapper import StatementTrapper
from py_hcl.core.type import HclType
from py_hcl.core.type.bundle import BundleT, Dir
from py_hcl.core.type.sint import SIntT
from py_hcl.core.type.uint import UIntT
from py_hcl.core.type.vector import VectorT
from py_hcl.utils import json_serialize


class ConnSide(Enum):
    UNKNOWN = 0
    LF = 1
    RT = 2
    BOTH = 3


@json_serialize
class Connect(object):
    def __init__(self, left, right):
        self.stmt_type = 'connect'
        self.left_expr_id = left.id
        self.right_expr_id = right.id


connector = op_register('<<=')


@connector(UIntT, UIntT)
def _(left, right):
    check_connect_dir(left, right)

    if left.hcl_type.width < right.hcl_type.width:
        msg = 'connect(): connecting {} to {} will truncate the bits'.format(
            right.hcl_type, left.hcl_type)
        logging.warning(msg)
        right = right[left.hcl_type.width - 1:0]

    if left.hcl_type.width > right.hcl_type.width:
        right = op_apply('extend')(right, left.hcl_type.width)

    assert left.hcl_type.width == right.hcl_type.width
    StatementTrapper.track(Connect(left, right))
    return left


@connector(SIntT, SIntT)
def _(left, right):
    check_connect_dir(left, right)

    if left.hcl_type.width < right.hcl_type.width:
        logging.warning(
            'connect(): connecting {} to {} will truncate the bits'.format(
                right.hcl_type, left.hcl_type
            ))
        right = right[left.hcl_type.width - 1:0].to_sint()

    if left.hcl_type.width > right.hcl_type.width:
        right = op_apply('extend')(right, left.hcl_type.width)

    assert left.hcl_type.width == right.hcl_type.width
    StatementTrapper.track(Connect(left, right))
    return left


@connector(UIntT, SIntT)
def _(left, right):
    msg = 'connect(): connecting SInt to UInt, an auto-conversion will occur'
    logging.warning(msg)

    if left.hcl_type.width < right.hcl_type.width:
        logging.warning(
            'connect(): connecting {} to {} will truncate the bits'.format(
                right.hcl_type, left.hcl_type
            ))
        return op_apply('<<=')(left, right[left.hcl_type.width - 1:0])

    return op_apply('<<=')(left, right.to_uint())


@connector(SIntT, UIntT)
def _(left, right):
    msg = 'connect(): connecting UInt to SInt, an auto-conversion will occur'
    logging.warning(msg)

    if left.hcl_type.width < right.hcl_type.width:
        logging.warning(
            'connect(): connecting {} to {} will truncate the bits'.format(
                right.hcl_type, left.hcl_type
            ))
        right = right[left.hcl_type.width - 1:0]

    return op_apply('<<=')(left, right.to_sint())


@connector(BundleT, BundleT)
def _(left, right):
    check_connect_dir(left, right)

    # TODO: Accurate Error Message
    dir_and_types = right.hcl_type.fields
    keys = dir_and_types.keys()
    assert keys == left.hcl_type.fields.keys()

    for k in keys:
        lf = op_apply('.')(left, k)
        rt = op_apply('.')(right, k)
        if dir_and_types[k]['dir'] == Dir.SRC:
            op_apply('<<=')(lf, rt)
        else:
            op_apply('<<=')(rt, lf)

    return left


@connector(VectorT, VectorT)
def _(left, right):
    check_connect_dir(left, right)

    # TODO: Accurate Error Message
    assert left.hcl_type.size == right.hcl_type.size

    for i in range(left.hcl_type.size):
        op_apply('<<=')(left[i], right[i])

    return left


@connector(HclType, HclType)
def _(_0, _1):
    raise StatementError.connect_type_error(_0, _1)


def check_connect_dir(left, right):
    # TODO: Accurate Error Message
    assert left.conn_side in (ConnSide.LF, ConnSide.BOTH)
    assert right.conn_side in (ConnSide.RT, ConnSide.BOTH)



from py_hcl.core.error import CoreError


def set_up():
    ExprError.append({
        'IOValueError': {
            'code': 200,
            'value': ExprError('io items should wrap with Input or Output')},
        'OpTypeError': {
            'code': 201,
            'value': ExprError('specified arguments contain unexpected types')
        }
    })


class ExprError(CoreError):
    @staticmethod
    def io_value(msg):
        return ExprError.err('IOValueError', msg)

    @staticmethod
    def op_type_err(op, *args):
        ts = ', '.join([type(a.hcl_type).__name__ for a in args])
        msg = '{}(): unsupported operand types: {}'.format(op, ts)
        return ExprError.err('OpTypeError', msg)


set_up()



import logging

from multipledispatch import dispatch

from ...stmt.connect import Connect
from ...type_measurer import equal

checker = dispatch


###############################################################
#                      TYPE CHECKERS                          #
###############################################################

@checker(Connect)
def check(connect: Connect):
    from .. import check_all_expr
    if not check_all_expr(connect.loc_ref, connect.expr_ref):
        logging.error("connect: lhs or rhs reference check failed")
        return False

    if not equal(connect.loc_ref.tpe, connect.expr_ref.tpe):
        logging.error("connect: type unmatched - {} & {}".format(
            connect.loc_ref.tpe, connect.expr_ref.tpe
        ))
        return False

    return True



from py_hcl.core.expr import HclExpr
from py_hcl.core.stmt.connect import ConnSide
from py_hcl.core.type import HclType
from py_hcl.utils import json_serialize


@json_serialize(json_fields=['hcl_type', 'conn_side'])
class Wire(HclExpr):
    def __init__(self, hcl_type: HclType):
        self.hcl_type = hcl_type
        self.conn_side = ConnSide.BOTH



from py_hcl.core.expr import ExprHolder
from py_hcl.core.expr.bundle_holder import BundleHolder
from py_hcl.core.expr.error import ExprError
from py_hcl.core.hcl_ops import op_register
from py_hcl.core.stmt.connect import ConnSide
from py_hcl.core.type import HclType
from py_hcl.core.type.bundle import BundleT, Dir
from py_hcl.utils import json_serialize

field_accessor = op_register('.')


@json_serialize
class FieldAccess(object):
    def __init__(self, expr, item):
        self.operation = "field_access"
        self.item = item
        self.ref_expr_id = expr.id


@field_accessor(BundleT)
def _(bd, item):
    # TODO: Accurate Error Message
    assert item in bd.hcl_type.fields
    if isinstance(bd, BundleHolder):
        return bd.assoc_value[item]

    # build connect side
    sd = bd.conn_side
    f = bd.hcl_type.fields[item]
    dr, tpe = f["dir"], f["hcl_type"]
    new_sd = build_new_sd(sd, dr)

    return ExprHolder(tpe, new_sd, FieldAccess(bd, item))


@field_accessor(HclType)
def _(o, *_):
    raise ExprError.op_type_err('field_accessor', o)


def build_new_sd(sd: ConnSide, dr: Dir) -> ConnSide:
    if sd == ConnSide.BOTH:
        return ConnSide.BOTH
    if sd == ConnSide.RT and dr == dr.SINK:
        return ConnSide.LF
    if sd == ConnSide.LF and dr == dr.SRC:
        return ConnSide.LF
    return ConnSide.RT



from py_hcl.core.expr.lit_sint import SLiteral
from py_hcl.core.type.sint import SIntT


class _(object):
    def __call__(self, value: int) -> SLiteral:
        return SLiteral(value)

    @staticmethod
    def w(width: int) -> SIntT:
        return SIntT(width)


S = _()



import logging

from multipledispatch import dispatch

from ...shortcuts import uw
from ...stmt.conditionally import Conditionally
from ...type_measurer import equal

checker = dispatch


###############################################################
#                      TYPE CHECKERS                          #
###############################################################

@checker(Conditionally)
def check(cond: Conditionally):
    from .. import check_all_expr, check_all_stmt
    if not check_all_expr(cond.pred_ref):
        logging.error("conditionally: cond reference check failed")
        return False

    if not check_all_stmt(cond.seq, cond.alt):
        logging.error("conditionally: seq or alt statements check failed")
        return False

    if not equal(cond.pred_ref.tpe, uw(1)):
        logging.error("conditionally: cond reference type check failed")
        return False

    return True



from py_hcl.firrtl_ir.expr.prim_ops import Rem
from py_hcl.firrtl_ir.shortcuts import uw, sw, u, w, s
from py_hcl.firrtl_ir.type import UIntType, SIntType
from tests.test_firrtl_ir.utils import serialize_equal
from .helper import OpCase, basis_tester, \
    encounter_error_tester, type_wrong_cases_2_args_gen, min_width


def args(*arg_types):
    class C:
        @staticmethod
        def tpe(res_type):
            return OpCase(Rem).arg_types(*arg_types).res_type(res_type)

    return C


rem_basis_cases = [
    args(UIntType, UIntType).tpe(lambda x, y: uw(min_width(x, y))),
    args(SIntType, SIntType).tpe(lambda x, y: sw(min_width(x, y))),
]

rem_type_wrong_cases = type_wrong_cases_2_args_gen(Rem)

rem_width_wrong_cases = [
    args(UIntType, UIntType).tpe(lambda x, y: uw(min_width(x, y) + 1)),
    args(SIntType, SIntType).tpe(lambda x, y: sw(min_width(x, y) + 1)),
    args(UIntType, UIntType).tpe(lambda x, y: uw(min_width(x, y) - 1)),
    args(SIntType, SIntType).tpe(lambda x, y: sw(min_width(x, y) - 1)),
]


def test_rem():
    basis_tester(rem_basis_cases)
    encounter_error_tester(rem_type_wrong_cases)
    encounter_error_tester(rem_width_wrong_cases)
    serialize_equal(Rem([u(20, w(5)), u(15, w(4))], uw(4)),
                    'rem(UInt<5>("h14"), UInt<4>("hf"))')
    serialize_equal(Rem([s(-20, w(6)), s(-15, w(5))], sw(5)),
                    'rem(SInt<6>("h-14"), SInt<5>("h-f"))')



from ..utils import serialize_str
from . import Expression


class Reference(Expression):
    def __init__(self, name, tpe):
        self.name = name
        self.tpe = tpe

    def serialize(self, output):
        output.write(serialize_str(self.name))



from py_hcl.firrtl_ir.expr.prim_ops import AsUInt
from py_hcl.firrtl_ir.shortcuts import uw, sw, u, w, s, n
from py_hcl.firrtl_ir.type import UIntType, SIntType, ClockType
from tests.test_firrtl_ir.utils import serialize_equal
from .helper import OpCase, basis_tester, \
    encounter_error_tester, type_wrong_cases_1_arg_gen, width


def args(*arg_types):
    class C:
        @staticmethod
        def tpe(res_type):
            return OpCase(AsUInt).arg_types(*arg_types).res_type(res_type)

    return C


asuint_basis_cases = [
    args(UIntType).tpe(lambda x: uw(width(x))),
    args(SIntType).tpe(lambda x: uw(width(x))),
    args(ClockType).tpe(lambda x: uw(1)),
]

asuint_type_wrong_cases = type_wrong_cases_1_arg_gen(AsUInt) + [
    args(UIntType).tpe(lambda x: sw(width(x))),
    args(SIntType).tpe(lambda x: sw(width(x))),
    args(ClockType).tpe(lambda x: sw(1)),
]

asuint_width_wrong_cases = [
    args(UIntType).tpe(lambda x: uw(width(x) + 1)),
    args(SIntType).tpe(lambda x: uw(width(x) + 1)),
    args(UIntType).tpe(lambda x: uw(width(x) - 1)),
    args(SIntType).tpe(lambda x: uw(width(x) - 1)),
    args(ClockType).tpe(lambda x: uw(2)),
]


def test_asuint():
    basis_tester(asuint_basis_cases)
    encounter_error_tester(asuint_type_wrong_cases)
    encounter_error_tester(asuint_width_wrong_cases)
    serialize_equal(AsUInt(u(20, w(5)), uw(5)),
                    'asUInt(UInt<5>("h14"))')
    serialize_equal(AsUInt(s(-20, w(6)), uw(5)),
                    'asUInt(SInt<6>("h-14"))')
    serialize_equal(AsUInt(n("clock", ClockType()), uw(1)),
                    'asUInt(clock)')



from ...utils import serialize_str
from .. import Statement


class DefWire(Statement):
    def __init__(self, name, tpe):
        self.name = name
        self.tpe = tpe

    def serialize_stmt(self, output, indent):
        output.write(b"wire ")
        output.write(serialize_str(self.name))
        output.write(b" : ")
        self.tpe.serialize(output)



from py_hcl.firrtl_ir.expr.prim_ops import Dshl
from py_hcl.firrtl_ir.shortcuts import uw, sw, u, w, s
from py_hcl.firrtl_ir.type import UIntType, SIntType
from tests.test_firrtl_ir.utils import serialize_equal
from .helper import OpCase, basis_tester, \
    encounter_error_tester, type_wrong_cases_2_args_gen, width


def args(*arg_types):
    class C:
        @staticmethod
        def tpe(res_type):
            return OpCase(Dshl).arg_types(*arg_types).res_type(res_type)

    return C


dshl_basis_cases = [
    args(UIntType, UIntType).tpe(
        lambda x, y: uw(width(x) + pow(2, width(y)) - 1)),
    args(SIntType, UIntType).tpe(
        lambda x, y: sw(width(x) + pow(2, width(y)) - 1)),
]

dshl_type_wrong_cases = type_wrong_cases_2_args_gen(Dshl)

dshl_width_wrong_cases = [
    args(UIntType, UIntType).tpe(
        lambda x, y: uw(width(x) + pow(2, width(y)))),
    args(SIntType, UIntType).tpe(
        lambda x, y: sw(width(x) + pow(2, width(y)))),
    args(UIntType, UIntType).tpe(
        lambda x, y: uw(width(x) + pow(2, width(y)) - 2)),
    args(SIntType, UIntType).tpe(
        lambda x, y: sw(width(x) + pow(2, width(y)) - 2)),
]


def test_dshl():
    basis_tester(dshl_basis_cases)
    encounter_error_tester(dshl_type_wrong_cases)
    encounter_error_tester(dshl_width_wrong_cases)
    serialize_equal(Dshl([u(20, w(5)), u(15, w(4))], uw(20)),
                    'dshl(UInt<5>("h14"), UInt<4>("hf"))')
    serialize_equal(Dshl([s(-20, w(6)), u(15, w(4))], uw(21)),
                    'dshl(SInt<6>("h-14"), UInt<4>("hf"))')



from py_hcl.utils import json_serialize


@json_serialize
class PackedModule(object):
    def __init__(self, name, named_expr_chain, statement_chain):
        self.name = name
        self.named_expr_chain = named_expr_chain
        self.statement_chain = statement_chain



from typing import Optional

from py_hcl.core.stmt import ClusterStatement
from py_hcl.utils import json_serialize


@json_serialize
class StmtHolder(object):
    def __init__(self, module_name: str, top_statement: ClusterStatement):
        self.module_name = module_name
        self.top_statement = top_statement


@json_serialize
class StmtNode(object):
    def __init__(self, stmt_holder: StmtHolder,
                 next_node: Optional["StmtNode"]):
        self.stmt_holder = stmt_holder
        if next_node:
            self.next_node = next_node


@json_serialize
class StmtChain(object):
    def __init__(self, stmt_chain_head: StmtNode):
        self.stmt_chain_head = stmt_chain_head



from py_hcl.core.expr.lit_uint import ULiteral
from py_hcl.core.type.uint import UIntT


class _(object):
    def __call__(self, value: int) -> ULiteral:
        return ULiteral(value)

    @staticmethod
    def w(width: int) -> UIntT:
        return UIntT(width)


U = _()
Bool = U.w(1)



from py_hcl.core.expr import HclExpr
from py_hcl.core.type.bundle import BundleT


class BundleHolder(HclExpr):
    def __init__(self, hcl_type, conn_side, assoc_value):
        self.hcl_type = hcl_type
        self.conn_side = conn_side

        assert isinstance(hcl_type, BundleT)
        assert set(hcl_type.fields.keys()) == set(assoc_value.keys())
        self.assoc_value = assoc_value



from typing import List

from .inherit_list.named_expr import NamedExprChain, \
    NamedExprNode, NamedExprHolder
from .inherit_list.stmt_holder import StmtChain, \
    StmtHolder, StmtNode


def merge_expr(modules: List[type],
               expr_holder: NamedExprHolder) -> NamedExprChain:
    expr_list = None
    for m in modules[::-1]:
        h = m.packed_module.named_expr_chain \
            .named_expr_chain_head.named_expr_holder
        expr_list = NamedExprNode(h, expr_list)

    expr_list = NamedExprNode(expr_holder, expr_list)
    return NamedExprChain(expr_list)


def merge_statement(modules: List[type],
                    stmt_holder: StmtHolder) -> StmtChain:
    stmt_list = None
    for m in modules[::-1]:
        h = m.packed_module.statement_chain \
            .stmt_chain_head.stmt_holder
        stmt_list = StmtNode(h, stmt_list)

    stmt_list = StmtNode(stmt_holder, stmt_list)
    return StmtChain(stmt_list)



import logging

logging.basicConfig(format='%(asctime)s %(message)s')


class PyHclError(BaseException):
    errs = {}

    @classmethod
    def append(cls, errs):
        PyHclError.errs[cls] = errs

    @classmethod
    def err(cls, name, msg):
        e = PyHclError.errs[cls][name]
        ev = e['value']

        logging.error("[%d]: %s", e['code'], msg)
        return ev



from py_hcl.core.expr import HclExpr
from py_hcl.core.stmt.connect import ConnSide
from py_hcl.core.type.uint import UIntT
from py_hcl.utils import unsigned_num_bin_len


class ULiteral(HclExpr):
    def __init__(self, value: int):
        self.value = value

        w = unsigned_num_bin_len(value)
        self.hcl_type = UIntT(w)
        self.conn_side = ConnSide.RT



from py_hcl.firrtl_ir.type.field import Field
from py_hcl.firrtl_ir.type import UnknownType, ClockType, \
    UIntType, SIntType, VectorType, BundleType
from py_hcl.firrtl_ir.type.width import Width
from .utils import serialize_equal


def test_unknown_type():
    serialize_equal(UnknownType(), "?")


def test_clock_type():
    serialize_equal(ClockType(), "Clock")


def test_uint_type():
    serialize_equal(UIntType(Width(8)), "UInt<8>")
    serialize_equal(UIntType(Width(32)), "UInt<32>")


def test_sint_type():
    serialize_equal(SIntType(Width(8)), "SInt<8>")
    serialize_equal(SIntType(Width(32)), "SInt<32>")


def test_vector_type():
    vt = VectorType(SIntType(Width(8)), 16)
    serialize_equal(vt, "SInt<8>[16]")

    vt = VectorType(UIntType(Width(8)), 16)
    serialize_equal(vt, "UInt<8>[16]")

    vt = VectorType(vt, 32)
    serialize_equal(vt, "UInt<8>[16][32]")

    vt = VectorType(VectorType(VectorType(vt, 42), 7), 9)
    serialize_equal(vt, "UInt<8>[16][32][42][7][9]")


def test_bundle_type():
    bd = BundleType([
        Field("a", UIntType(Width(8))),
        Field("b", UIntType(Width(8))),
        Field("c", UIntType(Width(8)), True),
    ])
    serialize_equal(bd, "{a : UInt<8>, b : UInt<8>, flip c : UInt<8>}")

    vt = VectorType(UIntType(Width(8)), 16)
    bd = BundleType([
        Field("a", vt),
        Field("b", UIntType(Width(8)), True),
        Field("c", VectorType(vt, 32)),
    ])
    serialize_equal(
        bd, "{a : UInt<8>[16], flip b : UInt<8>, c : UInt<8>[16][32]}"
    )

    # TODO: Is it valid?
    bd = BundleType([
        Field("l1", BundleType([
            Field("l2", BundleType([
                Field("l3", UIntType(Width(8)), True)
            ])),
            Field("vt", vt),
        ]))
    ])
    serialize_equal(bd, "{l1 : {l2 : {flip l3 : UInt<8>}, vt : UInt<8>[16]}}")



from py_hcl.firrtl_ir.expr.accessor import SubIndex
from py_hcl.firrtl_ir.shortcuts import n, s, w, vec, uw
from py_hcl.firrtl_ir.stmt.defn.node import DefNode
from py_hcl.firrtl_ir.type_checker import check
from ...utils import serialize_stmt_equal


def test_node_basis():
    node = DefNode("n1", s(20, w(6)))
    assert check(node)
    serialize_stmt_equal(node, 'node n1 = SInt<6>("h14")')

    node = DefNode("n2", SubIndex(n("v", vec(uw(8), 10)), 7, uw(8)))
    assert check(node)
    serialize_stmt_equal(node, 'node n2 = v[7]')


def test_node_expr_wrong():
    node = DefNode("n1", s(20, w(5)))
    assert not check(node)

    node = DefNode("n2", SubIndex(n("v", vec(uw(8), 10)), 10, uw(8)))
    assert not check(node)



import pytest

from py_hcl.core.expr import HclExpr
from py_hcl.core.module_factory.error import ModuleError
from py_hcl.dsl.expr.io import IO
from py_hcl.dsl.module import Module


def test_module():
    class A(Module):
        io = IO()
        a = HclExpr()

    assert hasattr(A, "packed_module")
    assert len(A.packed_module.named_expr_chain.named_expr_chain_head
               .named_expr_holder.named_expression_table) == 2


def test_module_not_contains_io():
    with pytest.raises(ModuleError, match='^.*lack of io.*$'):
        class A(Module):
            b = HclExpr()



from py_hcl.core.expr import ExprHolder
from py_hcl.core.expr.error import ExprError
from py_hcl.core.expr.utils import assert_right_side
from py_hcl.core.expr.vec_holder import VecHolder
from py_hcl.core.hcl_ops import op_register
from py_hcl.core.stmt.connect import ConnSide
from py_hcl.core.type import HclType
from py_hcl.core.type.sint import SIntT
from py_hcl.core.type.uint import UIntT
from py_hcl.core.type.vector import VectorT
from py_hcl.utils import json_serialize

slice_ = op_register('[i:j]')


@json_serialize
class Bits(object):
    def __init__(self, expr, high, low):
        self.operation = 'bits'
        self.high = high
        self.low = low
        self.ref_expr_id = expr.id


@slice_(UIntT)
@assert_right_side
def _(uint, high: int, low: int):
    check_bit_width(uint, high, low)
    t = UIntT(high - low + 1)
    return ExprHolder(t, ConnSide.RT, Bits(uint, high, low))


@slice_(SIntT)
@assert_right_side
def _(sint, high: int, low: int):
    check_bit_width(sint, high, low)
    t = UIntT(high - low + 1)
    return ExprHolder(t, ConnSide.RT, Bits(sint, high, low))


@slice_(VectorT)
def _(vec, low: int, high: int):
    check_vec_size(vec, low, high)

    if isinstance(vec, VecHolder):
        values = vec.assoc_value[low: high]
    else:
        values = [vec[i] for i in range(low, high, 1)]

    v_type = VectorT(vec.hcl_type.inner_type, high - low)
    return VecHolder(v_type, vec.conn_side, values)


@slice_(HclType)
def _(_0, *_):
    ExprError.op_type_err('slice', _0)


def check_bit_width(uint, high, low):
    w = uint.hcl_type.width
    # TODO: Accurate Error Message
    assert w > high >= low >= 0


def check_vec_size(vec, low, high):
    s = vec.hcl_type.size
    # TODO: Accurate Error Message
    assert 0 <= low < high <= s



from ...utils import serialize_str
from .. import Statement


class DefCircuit(Statement):
    def __init__(self, main, def_modules):
        self.main = main
        self.def_modules = def_modules

    def serialize_stmt(self, output, indent=0):
        output.write(b"circuit ")
        output.write(serialize_str(self.main))
        output.write(b" :\n")
        indent += 1
        for def_module in self.def_modules:
            output.write(serialize_str("  " * indent))
            def_module.serialize_stmt(output, indent)
            output.write(b"\n\n")



from py_hcl.core import install_ops
from py_hcl.core.expr.io import io_extend
from py_hcl.core.module.meta_module import MetaModule

install_ops()


class BaseModule(metaclass=MetaModule):
    io = io_extend(tuple())({})



from py_hcl.core.expr import ExprHolder
from py_hcl.core.expr.error import ExprError
from py_hcl.core.expr.utils import assert_right_side
from py_hcl.core.expr.vec_holder import VecHolder
from py_hcl.core.hcl_ops import op_register
from py_hcl.core.type import HclType
from py_hcl.core.type.sint import SIntT
from py_hcl.core.type.uint import UIntT
from py_hcl.core.type.vector import VectorT
from py_hcl.utils import json_serialize

index = op_register('[i]')


@json_serialize
class VecIndex(object):
    def __init__(self, expr, idx: int):
        self.operation = "vec_index"
        self.index = idx
        self.ref_expr_id = expr.id


@index(UIntT)
@assert_right_side
def _(uint, i: int):
    return uint[i:i]


@index(SIntT)
@assert_right_side
def _(sint, i: int):
    return sint[i:i]


@index(VectorT)
def _(vec, i: int):
    # TODO: Accurate Error Message
    assert i < vec.hcl_type.size
    if isinstance(vec, VecHolder):
        return vec.assoc_value[i]
    return ExprHolder(vec.hcl_type.inner_type, vec.conn_side, VecIndex(vec, i))


@index(HclType)
def _(_0, *_):
    raise ExprError.op_type_err('index', _0)



from py_hcl.firrtl_ir.shortcuts import n, uw, sw, u, w, s
from py_hcl.firrtl_ir.stmt.connect import Connect
from py_hcl.firrtl_ir.type_checker import check
from ..utils import serialize_stmt_equal


def test_connect_basis():
    cn = Connect(n("a", uw(8)), n("b", uw(8)))
    assert check(cn)
    serialize_stmt_equal(cn, "a <= b")

    cn = Connect(n("a", sw(8)), n("b", sw(8)))
    assert check(cn)
    serialize_stmt_equal(cn, "a <= b")

    cn = Connect(n("a", uw(8)), u(20, w(8)))
    assert check(cn)
    serialize_stmt_equal(cn, 'a <= UInt<8>("h14")')

    cn = Connect(n("a", sw(8)), s(-20, w(8)))
    assert check(cn)
    serialize_stmt_equal(cn, 'a <= SInt<8>("h-14")')


def test_connect_type_wrong():
    cn = Connect(n("a", uw(8)), n("b", sw(8)))
    assert not check(cn)

    cn = Connect(n("a", sw(8)), n("b", uw(8)))
    assert not check(cn)

    cn = Connect(n("a", uw(8)), s(20, w(8)))
    assert not check(cn)

    cn = Connect(n("a", sw(8)), n(-20, w(8)))
    assert not check(cn)



from . import Expression


class Mux(Expression):
    def __init__(self, cond, tval, fval, tpe):
        self.cond = cond
        self.tval = tval
        self.fval = fval
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"mux(")
        self.cond.serialize(output)
        output.write(b", ")
        self.tval.serialize(output)
        output.write(b", ")
        self.fval.serialize(output)
        output.write(b")")



from py_hcl.core.stmt import LineStatement, ClusterStatement
from .scope import ScopeManager, ScopeType


def set_up():
    ScopeManager.register_scope_expanding(
        StatementTrapper.on_scope_expanding)
    ScopeManager.register_scope_shrinking(
        StatementTrapper.on_scope_shrinking)
    ScopeManager.expand_scope(ScopeType.GROUND)


class StatementTrapper(object):
    trapped_stmts = [[
        # TOP SCOPE which contains all modules
    ]]

    @classmethod
    def trap(cls):
        ScopeManager.shrink_scope()

        assert len(cls.trapped_stmts) == 1
        ret = cls.trapped_stmts[0][-1]

        ScopeManager.expand_scope(ScopeType.GROUND)

        return ret  # TODO

    @classmethod
    def track(cls, statement):
        statement = LineStatement(
            ScopeManager.current_scope().scope_id,
            statement
        )
        cls.trapped_stmts[-1].append(statement)

    @classmethod
    def on_scope_expanding(cls, current_scope, next_scope):
        cls.trapped_stmts.append([])

    @classmethod
    def on_scope_shrinking(cls, current_scope, next_scope):
        stmts = cls.trapped_stmts.pop()
        cls.trapped_stmts[-1].append(ClusterStatement(current_scope, stmts))


set_up()



from py_hcl.firrtl_ir.shortcuts import vec, uw, bdl, n, u, w, s
from py_hcl.firrtl_ir.stmt.defn.memory import DefMemory, \
    DefMemWritePort, DefMemReadPort
from py_hcl.firrtl_ir.type import ClockType
from py_hcl.firrtl_ir.type_checker import check
from ...utils import serialize_stmt_equal


def test_memory_basis():
    mem = DefMemory("m", vec(uw(8), 10))
    assert check(mem)
    serialize_stmt_equal(mem, 'cmem m : UInt<8>[10]')

    mem = DefMemory("m", vec(bdl(a=(uw(8), False)), 10))
    assert check(mem)
    serialize_stmt_equal(mem, 'cmem m : {a : UInt<8>}[10]')


def test_memory_type_wrong():
    mem = DefMemory("m", bdl(a=(vec(uw(8), 10), False)))
    assert not check(mem)

    mem = DefMemory("m", uw(9))
    assert not check(mem)


def test_read_port_basis():
    mem_ref = n("m", vec(uw(8), 10))
    mr = DefMemReadPort("mr", mem_ref, u(2, w(8)), n("clock", ClockType()))
    assert check(mr)
    serialize_stmt_equal(mr, 'read mport mr = m[UInt<8>("h2")], clock')

    mem_ref = n("m", vec(bdl(a=(uw(8), False)), 10))
    mr = DefMemReadPort("mr", mem_ref, n("a", uw(2)), n("clock", ClockType()))
    assert check(mr)
    serialize_stmt_equal(mr, 'read mport mr = m[a], clock')


def test_read_port_clock_wrong():
    mem_ref = n("m", vec(uw(8), 10))
    mr = DefMemReadPort("mr", mem_ref, u(2, w(8)), n("clock", uw(1)))
    assert not check(mr)

    mem_ref = n("m", vec(bdl(a=(uw(8), False)), 10))
    mr = DefMemReadPort("mr", mem_ref, n("a", uw(2)), u(0, w(1)))
    assert not check(mr)


def test_read_port_index_wrong():
    mem_ref = n("m", vec(uw(8), 10))
    mr = DefMemReadPort("mr", mem_ref, s(2, w(8)), n("clock", ClockType()))
    assert not check(mr)

    mem_ref = n("m", vec(bdl(a=(uw(8), False)), 10))
    mr = DefMemReadPort("mr", mem_ref,
                        n("a", vec(uw(1), 10)), n("clock", ClockType()))
    assert not check(mr)


def test_read_port_mem_wrong():
    mem_ref = n("m", bdl(a=(vec(uw(8), 10), False)))
    mr = DefMemReadPort("mr", mem_ref, u(2, w(8)), n("clock", ClockType()))
    assert not check(mr)

    mem_ref = n("m", uw(9))
    mr = DefMemReadPort("mr", mem_ref, n("a", uw(2)), n("clock", ClockType()))
    assert not check(mr)


def test_write_port_basis():
    mem_ref = n("m", vec(uw(8), 10))
    mw = DefMemWritePort("mw", mem_ref, u(2, w(8)), n("clock", ClockType()))
    assert check(mw)
    serialize_stmt_equal(mw, 'write mport mw = m[UInt<8>("h2")], clock')

    mem_ref = n("m", vec(bdl(a=(uw(8), False)), 10))
    mw = DefMemWritePort("mw", mem_ref, n("a", uw(2)), n("clock", ClockType()))
    assert check(mw)
    serialize_stmt_equal(mw, 'write mport mw = m[a], clock')


def test_write_port_clock_wrong():
    mem_ref = n("m", vec(uw(8), 10))
    mw = DefMemWritePort("mw", mem_ref, u(2, w(8)), n("clock", uw(1)))
    assert not check(mw)

    mem_ref = n("m", vec(bdl(a=(uw(8), False)), 10))
    mw = DefMemWritePort("mw", mem_ref, n("a", uw(2)), u(0, w(1)))
    assert not check(mw)


def test_write_port_index_wrong():
    mem_ref = n("m", vec(uw(8), 10))
    mw = DefMemWritePort("mw", mem_ref, s(2, w(8)), n("clock", ClockType()))
    assert not check(mw)

    mem_ref = n("m", vec(bdl(a=(uw(8), False)), 10))
    mw = DefMemWritePort("mw", mem_ref,
                         n("a", vec(uw(1), 10)), n("clock", ClockType()))
    assert not check(mw)


def test_write_port_mem_wrong():
    mem_ref = n("m", bdl(a=(vec(uw(8), 10), False)))
    mw = DefMemWritePort("mw", mem_ref, u(2, w(8)), n("clock", ClockType()))
    assert not check(mw)

    mem_ref = n("m", uw(9))
    mw = DefMemWritePort("mw", mem_ref,
                         n("a", uw(2)), n("clock", ClockType()))
    assert not check(mw)



from ..utils import serialize_str
from . import Statement


class Block(Statement):
    def __init__(self, statements):
        self.statements = statements

    def serialize_stmt(self, output, indent):
        for stmt in self.statements[:-1]:
            stmt.serialize_stmt(output, indent)
            output.write(b"\n")
            output.write(serialize_str("  " * indent))
        self.statements[-1].serialize_stmt(output, indent)



from enum import Enum
from typing import Dict

from py_hcl.core.type import HclType
from py_hcl.core.type.wrapper import vec_wrap, bd_fld_wrap


class Dir(Enum):
    SRC = 1
    SINK = 2


@bd_fld_wrap
@vec_wrap
class BundleT(HclType):
    def __init__(self, fields: Dict[str, dict]):
        self.type = "bundle"
        self.fields = fields



import json
from enum import Enum
from functools import partial

from multipledispatch import dispatch


def signed_num_bin_len(num):
    return len("{:+b}".format(num))


def unsigned_num_bin_len(num):
    return len("{:b}".format(num))


def json_serialize(cls=None, json_fields=()):
    def rec(v):
        if hasattr(v, "json_obj"):
            return v.json_obj()
        elif isinstance(v, dict):
            return {k: rec(v) for k, v in v.items()}
        elif isinstance(v, (list, tuple)):
            return [rec(v) for v in v]
        elif isinstance(v, Enum):
            return v.name
        return v

    def serialize(self):
        return json.dumps(self.json_obj(), indent=2)

    def _(_cls, _json_fields):
        def js(self):
            if len(_json_fields) == 0:
                kv = vars(self)
            else:
                kv = {f: vars(self)[f] for f in _json_fields}

            return {k: rec(v) for k, v in kv.items()}

        _cls.json_obj = js
        _cls.__str__ = serialize
        return _cls

    if cls:
        return _(cls, json_fields)

    return partial(_, _json_fields=json_fields)


@dispatch()
def _fm(vd: dict):
    ls = ['{}: {}'.format(k, _fm(v)) for k, v in vd.items()]
    fs = _iter_repr(ls)
    return '{%s}' % (''.join(fs))


@dispatch()
def _fm(v: list):
    ls = [_fm(a) for a in v]
    fs = _iter_repr(ls)
    return '[%s]' % (''.join(fs))


@dispatch()
def _fm(v: tuple):
    ls = [_fm(a) for a in v]
    fs = _iter_repr(ls)
    return '(%s)' % (''.join(fs))


@dispatch()
def _fm(v: object):
    return str(v)


def _iter_repr(ls):
    if len(ls) <= 1:
        fs = ''.join(ls)
    else:
        fs = ''.join(['\n  {},'.format(_indent(l)) for l in ls]) + '\n'
    return fs


def _indent(s: str) -> str:
    s = s.split('\n')
    return '\n  '.join(s)



import copy

from multipledispatch import dispatch

from py_hcl.transformer.pyhcl_to_firrtl.context import Context
from py_hcl.transformer.pyhcl_to_firrtl.conv_port import ports_to_bundle_type
from py_hcl.transformer.pyhcl_to_firrtl.conv_type import convert_type
from py_hcl.transformer.pyhcl_to_firrtl.utils import build_io_name, get_io_obj
from py_hcl.core.expr import ExprHolder
from py_hcl.core.expr.add import Add as CAdd
from py_hcl.core.expr.and_ import And as CAnd
from py_hcl.core.expr.xor import Xor as CXor
from py_hcl.core.expr.or_ import Or as COr
from py_hcl.core.expr.convert import ToSInt, ToUInt
from py_hcl.core.expr.extend import Extend
from py_hcl.core.expr.field import FieldAccess
from py_hcl.core.expr.index import VecIndex
from py_hcl.core.expr.io import IO
from py_hcl.core.expr.lit_sint import SLiteral
from py_hcl.core.expr.lit_uint import ULiteral
from py_hcl.core.expr.mod_inst import ModuleInst
from py_hcl.core.expr.slice import Bits as CBits
from py_hcl.core.expr.wire import Wire
from py_hcl.firrtl_ir.expr.accessor import SubField, SubIndex
from py_hcl.firrtl_ir.expr.literal import UIntLiteral, SIntLiteral
from py_hcl.firrtl_ir.expr.prim_ops import Add, Bits, AsSInt, AsUInt, \
    And, Xor, Or
from py_hcl.firrtl_ir.expr.reference import Reference
from py_hcl.firrtl_ir.stmt.connect import Connect
from py_hcl.firrtl_ir.stmt.defn.instance import DefInstance
from py_hcl.firrtl_ir.stmt.defn.node import DefNode
from py_hcl.firrtl_ir.stmt.defn.wire import DefWire
from py_hcl.firrtl_ir.type import ClockType, UIntType
from py_hcl.firrtl_ir.type.width import Width


def convert_expr_by_id(expr_id: int):
    obj = Context.expr_table[expr_id]
    if id(obj) in Context.expr_obj_id_to_ref:
        return [], Context.expr_obj_id_to_ref[id(obj)]

    return convert_expr(obj)


@dispatch()
def convert_expr_op(expr_holder: ExprHolder, add: CAdd):
    l_stmts, l_ref = convert_expr_by_id(add.left_expr_id)
    r_stmts, r_ref = convert_expr_by_id(add.right_expr_id)
    name = NameGetter.get(expr_holder.id)
    typ = convert_type(expr_holder.hcl_type)
    stmt, ref = save_node_ref(Add([l_ref, r_ref], typ),
                              name, typ, id(expr_holder))
    return [*l_stmts, *r_stmts, stmt], ref


@dispatch()
def convert_expr_op(expr_holder: ExprHolder, and_: CAnd):
    l_stmts, l_ref = convert_expr_by_id(and_.left_expr_id)
    r_stmts, r_ref = convert_expr_by_id(and_.right_expr_id)
    name = NameGetter.get(expr_holder.id)
    typ = convert_type(expr_holder.hcl_type)
    stmt, ref = save_node_ref(And([l_ref, r_ref], typ),
                              name, typ, id(expr_holder))
    return [*l_stmts, *r_stmts, stmt], ref


@dispatch()
def convert_expr_op(expr_holder: ExprHolder, xor: CXor):
    l_stmts, l_ref = convert_expr_by_id(xor.left_expr_id)
    r_stmts, r_ref = convert_expr_by_id(xor.right_expr_id)
    name = NameGetter.get(expr_holder.id)
    typ = convert_type(expr_holder.hcl_type)
    stmt, ref = save_node_ref(Xor([l_ref, r_ref], typ),
                              name, typ, id(expr_holder))
    return [*l_stmts, *r_stmts, stmt], ref


@dispatch()
def convert_expr_op(expr_holder: ExprHolder, or_: COr):
    l_stmts, l_ref = convert_expr_by_id(or_.left_expr_id)
    r_stmts, r_ref = convert_expr_by_id(or_.right_expr_id)
    name = NameGetter.get(expr_holder.id)
    typ = convert_type(expr_holder.hcl_type)
    stmt, ref = save_node_ref(Or([l_ref, r_ref], typ),
                              name, typ, id(expr_holder))
    return [*l_stmts, *r_stmts, stmt], ref


@dispatch()
def convert_expr_op(expr_holder: ExprHolder, bs: CBits):
    stmts, v_ref = convert_expr_by_id(bs.ref_expr_id)
    name = NameGetter.get(expr_holder.id)
    typ = convert_type(expr_holder.hcl_type)
    stmt, ref = save_node_ref(Bits(v_ref, [bs.high, bs.low], typ),
                              name, typ, id(expr_holder))
    return [*stmts, stmt], ref


@dispatch()
def convert_expr_op(expr_holder: ExprHolder, ts: ToSInt):
    stmts, v_ref = convert_expr_by_id(ts.ref_expr_id)
    name = NameGetter.get(expr_holder.id)
    typ = convert_type(expr_holder.hcl_type)
    stmt, ref = save_node_ref(AsSInt(v_ref, typ),
                              name, typ, id(expr_holder))
    return [*stmts, stmt], ref


@dispatch()
def convert_expr_op(expr_holder: ExprHolder, tu: ToUInt):
    stmts, v_ref = convert_expr_by_id(tu.ref_expr_id)
    name = NameGetter.get(expr_holder.id)
    typ = convert_type(expr_holder.hcl_type)
    stmt, ref = save_node_ref(AsUInt(v_ref, typ),
                              name, typ, id(expr_holder))
    return [*stmts, stmt], ref


@dispatch()
def convert_expr_op(expr_holder: ExprHolder, et: Extend):
    stmts, v_ref = convert_expr_by_id(et.ref_expr_id)
    typ = convert_type(expr_holder.hcl_type)
    ref = copy.copy(v_ref)
    ref.tpe = typ
    Context.expr_obj_id_to_ref[id(expr_holder)] = ref
    return stmts, ref


@dispatch()
def convert_expr_op(expr_holder: ExprHolder, vi: VecIndex):
    stmts, v_ref = convert_expr_by_id(vi.ref_expr_id)
    ref = SubIndex(v_ref, vi.index, convert_type(expr_holder.hcl_type))
    Context.expr_obj_id_to_ref[id(expr_holder)] = ref
    return stmts, ref


@dispatch()
def convert_expr_op(expr_holder: ExprHolder, fa: FieldAccess):
    typ = convert_type(expr_holder.hcl_type)
    obj = Context.expr_table[fa.ref_expr_id]

    def fetch_current_io_holder(obj):
        current_node = obj.io_chain_head
        while True:
            if fa.item in current_node.io_holder.named_ports:
                return current_node.io_holder
            current_node = current_node.next_node

    if isinstance(obj, IO):
        io_holder = fetch_current_io_holder(obj)
        name = build_io_name(io_holder.module_name, fa.item)
        ref = Reference(name, typ)
        Context.expr_obj_id_to_ref[id(expr_holder)] = ref
        return [], ref

    elif isinstance(obj, ModuleInst):
        stmts, b_ref = convert_expr_by_id(obj.id)
        io_holder = fetch_current_io_holder(get_io_obj(obj.packed_module))
        name = build_io_name(io_holder.module_name, fa.item)
        ref = SubField(b_ref, name, typ)
        Context.expr_obj_id_to_ref[id(expr_holder)] = ref
        return stmts, ref

    else:
        stmts, src_ref = convert_expr_by_id(fa.ref_expr_id)
        ref = SubField(src_ref, fa.item, typ)
        Context.expr_obj_id_to_ref[id(expr_holder)] = ref
        return stmts, ref


@dispatch()
def convert_expr(expr_holder: ExprHolder):
    return convert_expr_op(expr_holder, expr_holder.op_node)


@dispatch()
def convert_expr(slit: SLiteral):
    ft = SIntLiteral(slit.value, convert_type(slit.hcl_type).width)
    Context.expr_obj_id_to_ref[id(slit)] = ft
    return [], ft


@dispatch()
def convert_expr(ulit: ULiteral):
    ft = UIntLiteral(ulit.value, convert_type(ulit.hcl_type).width)
    Context.expr_obj_id_to_ref[id(ulit)] = ft
    return [], ft


@dispatch()
def convert_expr(wire: Wire):
    name = NameGetter.get(wire.id)
    typ = convert_type(wire.hcl_type)

    stmt = DefWire(name, typ)
    ref = Reference(name, typ)
    Context.expr_obj_id_to_ref[id(wire)] = ref
    return [stmt], ref


@dispatch()
def convert_expr(mi: ModuleInst):
    if mi.module_name not in Context.modules:
        from .conv_module import convert_module
        convert_module(mi.packed_module)
    module = Context.modules[mi.module_name]
    name = NameGetter.get(mi.id)
    ref = Reference(name, ports_to_bundle_type(module.ports))
    stmts = [DefInstance(name, mi.module_name),
             Connect(SubField(ref, 'clock', ClockType()),
                     Reference('clock', ClockType())),
             Connect(SubField(ref, 'reset', UIntType(Width(1))),
                     Reference('reset', UIntType(Width(1))))]
    Context.expr_obj_id_to_ref[id(mi)] = ref
    return stmts, ref


class NameGetter(object):
    cnt = -1

    @classmethod
    def get(cls, expr_id: int):
        try:
            return Context.expr_id_to_name[expr_id]
        except KeyError:
            cls.cnt += 1
            return "_T_" + str(cls.cnt)


def save_node_ref(op_ir, name, tpe, obj_id):
    stmt = DefNode(name, op_ir)
    ref = Reference(name, tpe)
    Context.expr_obj_id_to_ref[obj_id] = ref
    return stmt, ref



from py_hcl.core.expr import HclExpr
from py_hcl.core.type.vector import VectorT


class VecHolder(HclExpr):
    def __init__(self, hcl_type, conn_side, assoc_value):
        self.hcl_type = hcl_type
        self.conn_side = conn_side

        assert isinstance(hcl_type, VectorT)
        assert hcl_type.size == len(assoc_value)
        self.assoc_value = assoc_value



from py_hcl.core.expr import ExprHolder
from py_hcl.core.expr.utils import assert_right_side
from py_hcl.core.hcl_ops import op_register
from py_hcl.core.stmt.connect import ConnSide
from py_hcl.core.type.sint import SIntT
from py_hcl.core.type.uint import UIntT
from py_hcl.utils import json_serialize

extend = op_register('extend')


@json_serialize
class Extend(object):
    def __init__(self, expr):
        self.operation = 'extend'
        self.ref_expr_id = expr.id


@extend(UIntT)
@assert_right_side
def _(uint, size):
    return ExprHolder(UIntT(size), ConnSide.RT, Extend(uint))


@extend(SIntT)
@assert_right_side
def _(sint, size):
    return ExprHolder(SIntT(size), ConnSide.RT, Extend(sint))



from typing import Union

import py_hcl.core.expr.io as cio
from py_hcl.core.module.base_module import BaseModule
from py_hcl.core.type import HclType


def IO(**named_ports: Union[cio.Input, cio.Output]) -> cio.IO:
    return cio.io_extend((BaseModule,))(named_ports)


def Input(hcl_type: HclType) -> cio.Input:
    return cio.Input(hcl_type)


def Output(hcl_type: HclType) -> cio.Output:
    return cio.Output(hcl_type)


def io_extend(*modules: type):
    def _(**named_ports: Union[cio.Input, cio.Output]):
        return cio.io_extend(modules)(named_ports)

    return _



from py_hcl.core.type import HclType
from py_hcl.core.type.wrapper import bd_fld_wrap, vec_wrap
from py_hcl.utils import signed_num_bin_len


@bd_fld_wrap
@vec_wrap
class SIntT(HclType):
    def __init__(self, width):
        self.type = "sint"
        self.width = width

    def __call__(self, value: int):
        from py_hcl.core.expr.lit_sint import SLiteral

        # TODO: Accurate Error Message
        assert signed_num_bin_len(value) <= self.width
        u = SLiteral(value)
        u.hcl_type = SIntT(self.width)
        return u



from py_hcl.transformer.pyhcl_to_firrtl.context import Context
from py_hcl.transformer.pyhcl_to_firrtl.conv_module import convert_module
from py_hcl.core.module.packed_module import PackedModule
from py_hcl.firrtl_ir.stmt.defn.circuit import DefCircuit
from py_hcl.firrtl_ir.type_checker import check


def convert(packed_module: PackedModule):
    convert_module(packed_module)
    modules = list(Context.modules.values())
    Context.modules.clear()
    Context.expr_obj_id_to_ref.clear()
    Context.expr_id_to_name.clear()
    cir = DefCircuit(packed_module.name, modules)
    assert check(cir)
    return cir



from py_hcl.firrtl_ir.stmt.empty import EmptyStmt
from py_hcl.firrtl_ir.type_checker import check
from ..utils import serialize_stmt_equal


def test_empty_serialize():
    serialize_stmt_equal(EmptyStmt(), "skip")
    assert check(EmptyStmt())



from py_hcl.firrtl_ir.stmt.defn.instance import DefInstance
from py_hcl.firrtl_ir.type_checker import check
from ...utils import serialize_stmt_equal


def test_instance_basis():
    i1 = DefInstance("i1", "ALU")
    assert check(i1)
    serialize_stmt_equal(i1, 'inst i1 of ALU')



from py_hcl.core.module.packed_module import PackedModule
from py_hcl.core.module_factory.inherit_list.named_expr import NamedExprHolder
from py_hcl.core.module_factory.inherit_list.stmt_holder import StmtHolder
from py_hcl.core.stmt_factory.trapper import StatementTrapper
from py_hcl.core.utils import module_inherit_mro
from . import extractor
from . import merger
from ..stmt import ClusterStatement, ConditionStatement
from ..stmt_factory.scope import ScopeType


def pack(bases, dct, name) -> PackedModule:
    raw_expr = extractor.extract(dct)
    raw_statement = normalize_conditional_branch(StatementTrapper.trap())

    named_expr_chain, statement_chain = \
        handle_inherit(bases, raw_expr, raw_statement, name)

    res = PackedModule(name, named_expr_chain, statement_chain)
    return res


def handle_inherit(bases, named_expression, top_statement, name):
    modules = module_inherit_mro(bases)

    named_expr_chain = \
        merger.merge_expr(modules, NamedExprHolder(name, named_expression))

    statement_chain = \
        merger.merge_statement(modules, StmtHolder(name, top_statement))

    return named_expr_chain, statement_chain


def normalize_conditional_branch(cluster_statement: ClusterStatement):
    new_statements = []

    waiting_when = None
    for s in cluster_statement.statements:
        if s.stmt_class == 'line':
            new_statements.append(s)
            continue

        normalized = normalize_conditional_branch(s)
        if s.scope_info.scope_type == ScopeType.WHEN:
            cond_id = normalized.scope_info.tag_object.cond_expr_id
            cs = ConditionStatement(cond_id, normalized.statements, None)
            new_statements.append(cs)
            waiting_when = cs

        elif s.scope_info.scope_type == ScopeType.ELSE_WHEN:
            cond_id = normalized.scope_info.tag_object.cond_expr_id
            cs = ConditionStatement(cond_id, normalized.statements, None)
            waiting_when.alt_stmts = [cs]
            waiting_when = cs

        else:
            waiting_when.alt_stmts = normalized.statements

    cluster_statement.statements = new_statements
    return cluster_statement



from ..type import UIntType, SIntType


def type_in(obj, *types):
    for t in types:
        if isinstance(obj, t):
            return True
    return False


def all_the_same(*objects):
    t = objects[0]
    for o in objects[1:]:
        if o != t:
            return False
    return True


def check_all_same_uint_sint(*types):
    for t in types:
        if not type_in(t, UIntType, SIntType):
            return False

    return all_the_same(*list(map(type, types)))



from multipledispatch import dispatch
import logging

from ...stmt.block import Block

checker = dispatch


###############################################################
#                      TYPE CHECKERS                          #
###############################################################

@checker(Block)
def check(block: Block):
    from .. import check_all_stmt
    if not check_all_stmt(*block.statements):
        logging.error("block: statements check failed")
        return False

    if len(block.statements) == 0:
        logging.error("block: number of statements is 0")
        return False

    return True



from py_hcl.firrtl_ir.shortcuts import n, uw, u, w
from py_hcl.firrtl_ir.stmt.block import Block
from py_hcl.firrtl_ir.stmt.conditionally import Conditionally
from py_hcl.firrtl_ir.stmt.connect import Connect
from py_hcl.firrtl_ir.stmt.defn.circuit import DefCircuit
from py_hcl.firrtl_ir.stmt.defn.module import DefModule, OutputPort, InputPort
from py_hcl.firrtl_ir.stmt.defn.node import DefNode
from py_hcl.firrtl_ir.stmt.empty import EmptyStmt
from py_hcl.firrtl_ir.type_checker import check
from tests.test_firrtl_ir.utils import serialize_stmt_equal


def test_circuit_basis():
    m1 = DefModule("m1", [OutputPort("p", uw(8))],
                   Connect(n("p", uw(8)), u(2, w(8))))

    m2 = DefModule("m2", [InputPort("b", uw(8)),
                          OutputPort("a", uw(8))],
                   Block([DefNode("n", u(1, w(1))),
                          Conditionally(n("n", uw(1)),
                                        EmptyStmt(),
                                        Connect(n("a", uw(8)), n("b", uw(8))))]
                         ))
    ct = DefCircuit("m1", [m1, m2])
    assert check(ct)
    serialize_stmt_equal(ct, 'circuit m1 :\n'
                             '  module m1 :\n'
                             '    output p : UInt<8>\n'
                             '\n'
                             '    p <= UInt<8>("h2")\n'
                             '\n'
                             '  module m2 :\n'
                             '    input b : UInt<8>\n'
                             '    output a : UInt<8>\n'
                             '\n'
                             '    node n = UInt<1>("h1")\n'
                             '    when n :\n'
                             '      skip\n'
                             '    else :\n'
                             '      a <= b\n'
                             '\n')


def test_circuit_module_not_exist():
    m1 = DefModule("m1", [OutputPort("p", uw(8))],
                   Connect(n("p", uw(8)), u(2, w(8))))

    m2 = DefModule("m2", [InputPort("b", uw(8)),
                          OutputPort("a", uw(8))],
                   Block([DefNode("n", u(1, w(1))),
                          Conditionally(n("n", uw(1)),
                                        EmptyStmt(),
                                        Connect(n("a", uw(8)), n("b", uw(8))))]
                         ))
    ct = DefCircuit("m3", [m1, m2])
    assert not check(ct)


def test_circuit_module_wrong():
    m1 = DefModule("m1", [OutputPort("p", uw(8))],
                   Connect(n("p", uw(8)), u(2, w(8))))
    m2 = DefModule("m2", [], Connect(n("p", uw(8)), u(2, w(8))))
    ct = DefCircuit("m1", [m1, m2])
    assert not check(ct)



from py_hcl.transformer.pyhcl_to_firrtl.convertor import convert


def compile_to_firrtl(module_class, path=None):
    m = convert(module_class.packed_module)

    if path is None:
        path = module_class.packed_module.name + ".fir"

    with open(path, 'wb') as f:
        m.serialize_stmt(f)



from py_hcl.core.expr import HclExpr
from py_hcl.core.stmt import ClusterStatement
from py_hcl.core.stmt.error import StatementError
from py_hcl.core.stmt_factory.scope import ScopeManager, ScopeType
from py_hcl.core.stmt_factory.trapper import StatementTrapper
from py_hcl.core.type.uint import UIntT
from py_hcl.utils import json_serialize


@json_serialize
class When(object):
    def __init__(self, cond_expr_id: int):
        self.cond_expr_id = cond_expr_id


@json_serialize
class ElseWhen(object):
    def __init__(self, cond_expr_id: int):
        self.cond_expr_id = cond_expr_id


def do_when_enter(cond_expr: HclExpr):
    check_bool_expr(cond_expr)

    w = When(cond_expr.id)
    ScopeManager.expand_scope(ScopeType.WHEN, w)


def do_when_exit():
    ScopeManager.shrink_scope()


def do_else_when_enter(cond_expr: HclExpr):
    check_bool_expr(cond_expr)
    check_branch_syntax()

    e = ElseWhen(cond_expr.id)
    ScopeManager.expand_scope(ScopeType.ELSE_WHEN, e)


def do_else_when_exit():
    ScopeManager.shrink_scope()


def do_otherwise_enter():
    check_branch_syntax()

    ScopeManager.expand_scope(ScopeType.OTHERWISE)


def do_otherwise_exit():
    ScopeManager.shrink_scope()


def check_bool_expr(cond_expr: HclExpr):
    if isinstance(cond_expr.hcl_type, UIntT) and cond_expr.hcl_type.width == 1:
        return
    raise StatementError.wrong_branch_syntax(
        'check_bool_expr(): '
        'expected bool-type expression')


def check_branch_syntax():
    check_exists_pre_stmts()
    check_exists_pre_when_block()
    check_correct_block_level()


def check_exists_pre_stmts():
    if len(StatementTrapper.trapped_stmts[-1]) == 0:
        raise StatementError.wrong_branch_syntax(
            'check_exists_pre_stmts(): '
            'expected when block')


def check_exists_pre_when_block():
    last_stmt = StatementTrapper.trapped_stmts[-1][-1]

    if isinstance(last_stmt, ClusterStatement):
        last_scope = last_stmt.scope_info
        last_scope_type = last_scope.scope_type

        when = last_scope_type == ScopeType.WHEN
        else_when = last_scope_type == ScopeType.ELSE_WHEN

        if when or else_when:
            return

    raise StatementError.wrong_branch_syntax(
        'check_exists_pre_when_block(): '
        'expected when block or else_when block')


def check_correct_block_level():
    last_stmt = StatementTrapper.trapped_stmts[-1][-1]

    if isinstance(last_stmt, ClusterStatement):
        last_scope = last_stmt.scope_info
        current_scope = ScopeManager.current_scope()
        last_scope_level = last_scope.scope_level
        current_scope_level = current_scope.scope_level
        if last_scope_level == current_scope_level + 1:
            return

    raise StatementError.wrong_branch_syntax(
        'check_correct_block_level(): '
        'branch block not matched')



from multipledispatch import dispatch
import logging

from .stmt.definition import check
from .stmt.block import check
from .stmt.conditionally import check
from ..stmt import Statement
from ..expr import Expression
from .stmt.connect import check
from ..stmt.empty import EmptyStmt
from ..expr.reference import Reference
from .expr.literal import check
from .expr.mux import check
from .expr.prim_ops import check
from .expr.accessor import check

checker = dispatch


@checker(EmptyStmt)
def check(_: EmptyStmt):
    return True


@checker(Reference)
def check(_: Reference):
    return True


@checker(object)
def check(_: object):
    logging.error("unsupported type: " + _.__class__.__name__)
    return False


def check_all_expr(*obj):
    return all(isinstance(o, Expression) and check(o) for o in obj)


def check_all_stmt(*obj):
    return all(isinstance(o, Statement) and check(o) for o in obj)



from py_hcl.firrtl_ir.shortcuts import n, s, w, vec, uw, sw, u
from py_hcl.firrtl_ir.stmt.defn.register import DefRegister, DefInitRegister
from py_hcl.firrtl_ir.type import ClockType
from py_hcl.firrtl_ir.type_checker import check
from ...utils import serialize_stmt_equal


def test_register_basis():
    r1 = DefRegister("r1", uw(8), n("clock", ClockType()))
    assert check(r1)
    serialize_stmt_equal(r1, 'reg r1 : UInt<8>, clock')

    r2 = DefRegister("r2", vec(uw(8), 10), n("clock", ClockType()))
    assert check(r2)
    serialize_stmt_equal(r2, 'reg r2 : UInt<8>[10], clock')


def test_register_clock_wrong():
    r1 = DefRegister("r1", uw(8), n("clock", uw(1)))
    assert not check(r1)

    r2 = DefRegister("r2", vec(uw(8), 10), n("clock", sw(1)))
    assert not check(r2)


def test_init_register_basis():
    r1 = DefInitRegister("r1", uw(8),
                         n("clock", ClockType()), n("r", uw(1)), u(5, w(8)))
    assert check(r1)
    serialize_stmt_equal(r1, 'reg r1 : UInt<8>, clock with :\n'
                             '  reset => (r, UInt<8>("h5"))')

    r2 = DefInitRegister("r2", sw(8),
                         n("clock", ClockType()), u(0, w(1)), s(5, w(8)))
    assert check(r2)
    serialize_stmt_equal(r2, 'reg r2 : SInt<8>, clock with :\n'
                             '  reset => (UInt<1>("h0"), SInt<8>("h5"))')


def test_init_register_clock_wrong():
    r1 = DefInitRegister("r1", uw(8),
                         n("clock", uw(1)), n("r", uw(1)), u(5, w(8)))
    assert not check(r1)

    r2 = DefInitRegister("r2", sw(8),
                         n("clock", sw(1)), u(0, w(1)), s(5, w(8)))
    assert not check(r2)


def test_init_register_reset_wrong():
    r1 = DefInitRegister("r1", uw(8),
                         n("clock", ClockType()), n("r", sw(1)), u(5, w(8)))
    assert not check(r1)

    r2 = DefInitRegister("r2", sw(8),
                         n("clock", ClockType()), s(0, w(1)), s(5, w(8)))
    assert not check(r2)


def test_init_register_type_not_match():
    r1 = DefInitRegister("r1", uw(8),
                         n("clock", ClockType()), n("r", uw(1)), s(5, w(8)))
    assert not check(r1)

    r2 = DefInitRegister("r2", uw(8),
                         n("clock", ClockType()), u(0, w(1)), s(5, w(8)))
    assert not check(r2)



from py_hcl.firrtl_ir.shortcuts import uw, bdl, sw
from py_hcl.firrtl_ir.stmt.defn.wire import DefWire
from py_hcl.firrtl_ir.type_checker import check
from ...utils import serialize_stmt_equal


def test_wire_basis():
    wire = DefWire("w1", uw(8))
    assert check(wire)
    serialize_stmt_equal(wire, 'wire w1 : UInt<8>')

    wire = DefWire("w2", bdl(a=(uw(8), True), b=(sw(8), False)))
    assert check(wire)
    serialize_stmt_equal(wire, 'wire w2 : {flip a : UInt<8>, b : SInt<8>}')



from ...utils import serialize_str
from .. import Statement


class DefInstance(Statement):
    def __init__(self, name, module_name):
        self.name = name
        self.module_name = module_name

    def serialize_stmt(self, output, indent):
        output.write(b"inst ")
        output.write(serialize_str(self.name))
        output.write(b" of ")
        output.write(serialize_str(self.module_name))



from multipledispatch import dispatch

from py_hcl.transformer.pyhcl_to_firrtl.conv_expr import convert_expr_by_id
from py_hcl.core.stmt import LineStatement, \
    ClusterStatement, ConditionStatement
from py_hcl.core.stmt.connect import Connect as HclConnect
from py_hcl.firrtl_ir.stmt.block import Block
from py_hcl.firrtl_ir.stmt.conditionally import Conditionally
from py_hcl.firrtl_ir.stmt.connect import Connect
from py_hcl.firrtl_ir.stmt.empty import EmptyStmt


@dispatch()
def convert_stmt(line: LineStatement):
    return convert_stmt(line.statement)


@dispatch()
def convert_stmt(cluster: ClusterStatement):
    return [ss for s in cluster.statements for ss in convert_stmt(s)]


@dispatch()
def convert_stmt(condition: ConditionStatement):
    cond_stmts, cond_ref = convert_expr_by_id(condition.seq_cond_id)
    seq = [ss for s in condition.seq_stmts for ss in convert_stmt(s)]
    alt = [EmptyStmt()] if condition.alt_stmts is None else \
        [ss for s in condition.alt_stmts for ss in convert_stmt(s)]

    conditionally = Conditionally(cond_ref, Block(seq), Block(alt))
    return [*cond_stmts, conditionally]


@dispatch()
def convert_stmt(connect: HclConnect):
    stmts_0, left_ref = convert_expr_by_id(connect.left_expr_id)
    stmts_1, right_ref = convert_expr_by_id(connect.right_expr_id)
    return [*stmts_0, *stmts_1, Connect(left_ref, right_ref)]



from multipledispatch import dispatch

from py_hcl.firrtl_ir.type.field import Field

measurer = dispatch


@measurer(Field, Field)
def equal(f1: Field, f2: Field):
    if f1.name != f2.name:
        return False
    if f1.is_flipped != f2.is_flipped:
        return False

    return equal(f1.tpe, f2.tpe)



from py_hcl.firrtl_ir.shortcuts import n, uw, sw, u, w
from py_hcl.firrtl_ir.stmt.block import Block
from py_hcl.firrtl_ir.stmt.conditionally import Conditionally
from py_hcl.firrtl_ir.stmt.connect import Connect
from py_hcl.firrtl_ir.stmt.empty import EmptyStmt
from py_hcl.firrtl_ir.type_checker import check
from ..utils import serialize_stmt_equal


def test_conditionally_basis():
    s1 = EmptyStmt()
    s2 = Connect(n("a", uw(8)), n("b", uw(8)))
    cn = Conditionally(n("a", uw(1)), s1, s2)
    assert check(cn)
    serialize_stmt_equal(cn, "when a :\n"
                             "  skip\n"
                             "else :\n"
                             "  a <= b")

    s1 = Block([
        Connect(n("a", uw(8)), n("b", uw(8))),
        Connect(n("c", sw(8)), n("d", sw(8))),
    ])
    s2 = EmptyStmt()
    cn = Conditionally(u(1, w(1)), s1, s2)
    assert check(cn)
    serialize_stmt_equal(
        cn, 'when UInt<1>("h1") :\n'
            '  a <= b\n'
            '  c <= d\n'
            'else :\n'
            '  skip')


def test_conditionally_type_wrong():
    s1 = EmptyStmt()
    s2 = Connect(n("a", uw(8)), n("b", uw(8)))
    cn = Conditionally(n("a", sw(1)), s1, s2)
    assert not check(cn)

    s1 = Block([
        Connect(n("a", uw(8)), n("b", uw(8))),
        Connect(n("c", sw(8)), n("d", sw(8))),
    ])
    s2 = EmptyStmt()
    cn = Conditionally(u(1, w(2)), s1, s2)
    assert not check(cn)



from multipledispatch import dispatch

from ...expr.mux import Mux
from ...shortcuts import uw
from ...type_measurer import equal

checker = dispatch


###############################################################
#                      TYPE CHECKERS                          #
###############################################################

@checker(Mux)
def check(mux: Mux):
    from .. import check_all_expr
    if not check_all_expr(mux.cond, mux.tval, mux.fval):
        return False
    if not equal(mux.cond.tpe, uw(1)):
        return False
    if not equal(mux.tval.tpe, mux.fval.tpe):
        return False
    if not equal(mux.tval.tpe, mux.tpe):
        return False
    return True



from multipledispatch import dispatch

from ..type import UnknownType, UIntType, SIntType, \
    ClockType, BundleType, VectorType

measurer = dispatch


@measurer(UnknownType, UnknownType)
def equal(_0: UnknownType, _1: UnknownType):
    return True


@measurer(UIntType, UIntType)
def equal(t1: UIntType, t2: UIntType):
    return equal(t1.width, t2.width)


@measurer(SIntType, SIntType)
def equal(t1: SIntType, t2: SIntType):
    return equal(t1.width, t2.width)


@measurer(ClockType, ClockType)
def equal(_0: ClockType, _1: ClockType):
    return True


@measurer(BundleType, BundleType)
def equal(t1: BundleType, t2: BundleType):
    if len(t1.fields) != len(t2.fields):
        return False
    for (a, b) in zip(t1.fields, t2.fields):
        if not equal(a, b):
            return False
    return True


@measurer(VectorType, VectorType)
def equal(t1: VectorType, t2: VectorType):
    if t1.size != t2.size:
        return False
    if not equal(t1.elem_type, t2.elem_type):
        return False
    return True



from multipledispatch import dispatch

from ..utils import type_in
from ...shortcuts import uw
from ...stmt.defn.circuit import DefCircuit
from ...stmt.defn.instance import DefInstance
from ...stmt.defn.memory import DefMemory, DefMemReadPort, DefMemWritePort
from ...stmt.defn.module import DefModule, InputPort, OutputPort, DefExtModule
from ...stmt.defn.node import DefNode
from ...stmt.defn.register import DefRegister, DefInitRegister
from ...stmt.defn.wire import DefWire
from ...type import VectorType, ClockType, UIntType
from ...type_measurer import equal

checker = dispatch


###############################################################
#                      TYPE CHECKERS                          #
###############################################################

@checker(DefWire)
def check(_: DefWire):
    return True


@checker(DefInstance)
def check(_: DefInstance):
    return True


@checker(DefRegister)
def check(reg: DefRegister):
    from ...type_checker import check_all_expr
    if not check_all_expr(reg.clock_ref):
        return False

    if not type_in(reg.clock_ref.tpe, ClockType):
        return False

    return True


@checker(DefInitRegister)
def check(reg: DefInitRegister):
    from ...type_checker import check_all_expr
    if not check_all_expr(reg.clock_ref, reg.reset_ref, reg.init_ref):
        return False

    if not type_in(reg.clock_ref.tpe, ClockType):
        return False

    if not equal(reg.reset_ref.tpe, uw(1)):
        return False

    if not equal(reg.init_ref.tpe, reg.tpe):
        return False

    return True


@checker(DefNode)
def check(node: DefNode):
    from ...type_checker import check_all_expr
    if not check_all_expr(node.expr_ref):
        return False

    return True


@checker(DefMemory)
def check(mem: DefMemory):
    if not type_in(mem.tpe, VectorType):
        return False

    return True


@checker(DefMemReadPort)
def check(mem_read: DefMemReadPort):
    from ...type_checker import check_all_expr
    if not check_all_expr(mem_read.mem_ref, mem_read.index_ref,
                          mem_read.clock_ref):
        return False

    if not type_in(mem_read.mem_ref.tpe, VectorType):
        return False

    if not type_in(mem_read.clock_ref.tpe, ClockType):
        return False

    if not type_in(mem_read.index_ref.tpe, UIntType):
        return False

    return True


@checker(DefMemWritePort)
def check(mem_write: DefMemWritePort):
    from ...type_checker import check_all_expr
    if not check_all_expr(mem_write.mem_ref, mem_write.index_ref,
                          mem_write.clock_ref):
        return False

    if not type_in(mem_write.mem_ref.tpe, VectorType):
        return False

    if not type_in(mem_write.clock_ref.tpe, ClockType):
        return False

    if not type_in(mem_write.index_ref.tpe, UIntType):
        return False

    return True


@checker(DefModule)
def check(mod: DefModule):
    from ...type_checker import check_all_stmt
    if not check_all_stmt(mod.body):
        return False

    if len(mod.ports) == 0:
        return False

    for port in mod.ports:
        if not type_in(port, InputPort, OutputPort):
            return False

    return True


@checker(DefExtModule)
def check(mod: DefExtModule):
    if len(mod.ports) == 0:
        return False

    for port in mod.ports:
        if not type_in(port, InputPort, OutputPort):
            return False

    return True


@checker(DefCircuit)
def check(circuit: DefCircuit):
    from ...type_checker import check_all_stmt
    if not check_all_stmt(*circuit.def_modules):
        return False

    name_found = False
    for mod in circuit.def_modules:
        if not type_in(mod, DefModule, DefExtModule):
            return False
        if mod.name == circuit.main:
            name_found = True

    if not name_found:
        return False

    return True



from py_hcl.core.error import CoreError


def set_up():
    ModuleError.append({
        'NotContainsIO': {
            'code': 100,
            'value': ModuleError('the module lack of io attribute')},
    })


class ModuleError(CoreError):
    @staticmethod
    def not_contains_io(msg):
        return ModuleError.err('NotContainsIO', msg)


set_up()



from py_hcl.core.expr import ExprHolder
from py_hcl.core.expr.bundle_holder import BundleHolder
from py_hcl.core.expr.error import ExprError
from py_hcl.core.expr.utils import assert_right_side
from py_hcl.core.expr.vec_holder import VecHolder
from py_hcl.core.hcl_ops import op_register
from py_hcl.core.stmt.connect import ConnSide
from py_hcl.core.type import HclType
from py_hcl.core.type.bundle import BundleT, Dir
from py_hcl.core.type.sint import SIntT
from py_hcl.core.type.uint import UIntT
from py_hcl.core.type.vector import VectorT
from py_hcl.utils import json_serialize


@json_serialize
class Or(object):
    def __init__(self, left, right):
        self.operation = 'or'
        self.left_expr_id = left.id
        self.right_expr_id = right.id


orer = op_register('|')


@orer(UIntT, UIntT)
@assert_right_side
def _(lf, rt):
    w = max(lf.hcl_type.width, rt.hcl_type.width)
    t = UIntT(w)
    return ExprHolder(t, ConnSide.RT, Or(lf, rt))


@orer(SIntT, SIntT)
@assert_right_side
def _(lf, rt):
    w = max(lf.hcl_type.width, rt.hcl_type.width)
    t = UIntT(w)
    return ExprHolder(t, ConnSide.RT, Or(lf, rt))


@orer(VectorT, VectorT)
@assert_right_side
def _(lf, rt):
    # TODO: Accurate Error Message
    assert lf.hcl_type.size == rt.hcl_type.size

    values = [lf[i] | rt[i] for i in range(lf.hcl_type.size)]
    v_type = VectorT(values[0].hcl_type, len(values))
    return VecHolder(v_type, ConnSide.RT, values)


@orer(BundleT, BundleT)
@assert_right_side
def _(lf, rt):
    # TODO: Accurate Error Message
    assert set(lf.hcl_type.fields.keys()) == set(rt.hcl_type.fields.keys())

    bd_type_fields = {}
    bd_values = {}
    for k in lf.hcl_type.fields.keys():
        res = getattr(lf, k) | getattr(rt, k)
        bd_type_fields[k] = {"dir": Dir.SRC, "hcl_type": res.hcl_type}
        bd_values[k] = res

    return BundleHolder(BundleT(bd_type_fields), ConnSide.RT, bd_values)


@orer(HclType, HclType)
def _(_0, _1):
    raise ExprError.op_type_err('or', _0, _1)



from py_hcl.firrtl_ir.expr.prim_ops import Dshr
from py_hcl.firrtl_ir.shortcuts import uw, sw, u, w, s
from py_hcl.firrtl_ir.type import UIntType, SIntType
from tests.test_firrtl_ir.utils import serialize_equal
from .helper import OpCase, basis_tester, \
    encounter_error_tester, type_wrong_cases_2_args_gen, width


def args(*arg_types):
    class C:
        @staticmethod
        def tpe(res_type):
            return OpCase(Dshr).arg_types(*arg_types).res_type(res_type)

    return C


dshr_basis_cases = [
    args(UIntType, UIntType).tpe(
        lambda x, y: uw(width(x))),
    args(SIntType, UIntType).tpe(
        lambda x, y: sw(width(x))),
]

dshr_type_wrong_cases = type_wrong_cases_2_args_gen(Dshr)

dshr_width_wrong_cases = [
    args(UIntType, UIntType).tpe(
        lambda x, y: uw(width(x) + 1)),
    args(SIntType, UIntType).tpe(
        lambda x, y: sw(width(x) + 1)),
    args(UIntType, UIntType).tpe(
        lambda x, y: uw(width(x) - 1)),
    args(SIntType, UIntType).tpe(
        lambda x, y: sw(width(x) - 1)),
]


def test_dshr():
    basis_tester(dshr_basis_cases)
    encounter_error_tester(dshr_type_wrong_cases)
    encounter_error_tester(dshr_width_wrong_cases)
    serialize_equal(Dshr([u(20, w(5)), u(15, w(4))], uw(5)),
                    'dshr(UInt<5>("h14"), UInt<4>("hf"))')
    serialize_equal(Dshr([s(-20, w(6)), u(15, w(4))], uw(6)),
                    'dshr(SInt<6>("h-14"), UInt<4>("hf"))')



from multipledispatch import Dispatcher

op_map = {
    '+': Dispatcher('+'),
    '&': Dispatcher('&'),
    '^': Dispatcher('&'),
    '|': Dispatcher('|'),
    '<<=': Dispatcher('<<='),
    '.': Dispatcher('.'),
    '[i]': Dispatcher('[i]'),
    '[i:j]': Dispatcher('[i:j]'),
    '[i:]': Dispatcher('[i:]'),
    '[i:j:k]': Dispatcher('[i:j:k]'),
    '[i::k]': Dispatcher('[i::k]'),
    'to_sint': Dispatcher('to_sint'),
    'to_uint': Dispatcher('to_uint'),
    'to_bool': Dispatcher('to_bool'),
    'extend': Dispatcher('extend'),
}


def op_register(operation):
    return op_map[operation].register


def op_apply(operation):
    def _(*objects):
        types = [type(o.hcl_type) for o in objects if hasattr(o, 'hcl_type')]
        func = op_map[operation].dispatch(*types)

        if func is not None:
            return func(*objects)

        msg = 'No matched functions for types {} while calling operation ' \
              '"{}"'.format([type(o).__name__ for o in objects], operation)
        raise NotImplementedError(msg)

    return _



from ..type import UIntType, SIntType
from . import Expression
from ..utils import serialize_str


class UIntLiteral(Expression):
    def __init__(self, value, width):
        self.value = value
        self.tpe = UIntType(width)

    def serialize(self, output):
        self.tpe.serialize(output)
        output.write(b'("')
        output.write(serialize_str("h" + hex(self.value).replace("0x", "")))
        output.write(b'")')


class SIntLiteral(Expression):
    def __init__(self, value, width):
        self.value = value
        self.tpe = SIntType(width)

    def serialize(self, output):
        self.tpe.serialize(output)
        output.write(b'("')
        output.write(serialize_str("h" + hex(self.value).replace("0x", "")))
        output.write(b'")')



from py_hcl.core.expr.mod_inst import con_module
from py_hcl.core.module_factory import packer
from py_hcl.core.module_factory.error import ModuleError


class MetaModule(type):
    def __init__(cls, name, bases, dct):
        if name == '_hcl_fake_module':
            return

        super().__init__(name, bases, dct)
        cls.__new__ = con_module

        name = fetch_module_name(name)

        check_io_exist(dct, name)
        dct["io"].io_chain_head.io_holder.module_name = name

        packed = packer.pack(bases, dct, name)
        cls.packed_module = packed


module_names = {}


def fetch_module_name(name: str) -> str:
    n = module_names.get(name, 0)
    module_names[name] = n + 1
    return name + ("" if n == 0 else str(n))


def check_io_exist(dct, name):
    if 'io' not in dct:
        raise ModuleError.not_contains_io(
            'module {} lack of io attribute'.format(name))



from py_hcl.firrtl_ir.expr.prim_ops import Shl
from py_hcl.firrtl_ir.shortcuts import sw, uw, u, w, s
from py_hcl.firrtl_ir.type import BundleType, SIntType, \
    UIntType, UnknownType, VectorType
from tests.test_firrtl_ir.utils import serialize_equal
from .helper import basis_tester, encounter_error_tester, OpCase, width


def args(*arg_types):
    class A:
        @staticmethod
        def const(*const_types):
            class B:
                @staticmethod
                def tpe(res_type):
                    return OpCase(Shl) \
                        .arg_types(*arg_types) \
                        .const_arg_types(*const_types) \
                        .res_type(res_type)

            return B

    return A


shl_basis_cases = [
    args(UIntType).const(int).tpe(lambda x, y: uw(width(x) + y)),
    args(SIntType).const(int).tpe(lambda x, y: sw(width(x) + y)),
]

shl_type_wrong_cases = [
    args(UnknownType).const(int).tpe(lambda x, y: uw(y)),
    args(VectorType).const(int).tpe(lambda x, y: uw(y)),
    args(BundleType).const(int).tpe(lambda x, y: uw(y)),
]

shl_width_wrong_cases = [
    args(UIntType).const(int).tpe(lambda x, y: uw(width(x) + y + 1)),
    args(SIntType).const(int).tpe(lambda x, y: sw(width(x) + y + 1)),
    args(UIntType).const(int).tpe(lambda x, y: uw(width(x) + y - 1)),
    args(SIntType).const(int).tpe(lambda x, y: sw(width(x) + y - 1)),
]


def test_shl():
    basis_tester(shl_basis_cases)
    encounter_error_tester(shl_type_wrong_cases)
    encounter_error_tester(shl_width_wrong_cases)
    serialize_equal(Shl(u(20, w(5)), 6, uw(11)),
                    'shl(UInt<5>("h14"), 6)')
    serialize_equal(Shl(s(-20, w(6)), 6, sw(12)),
                    'shl(SInt<6>("h-14"), 6)')



from multipledispatch import dispatch

from .type import equal
from .width import equal
from .field import equal

measurer = dispatch


@measurer(object, object)
def equal(_0: object, _1: object):
    return False



from py_hcl.core.type import HclType
from py_hcl.core.type.wrapper import bd_fld_wrap


@bd_fld_wrap
class VectorT(HclType):
    def __init__(self, inner_type: HclType, size: int):
        # TODO: Accurate Error Message
        assert size > 0
        self.type = "vector"
        self.size = size
        self.inner_type = inner_type



from py_hcl.core.stmt.connect import ConnSide


def assert_right_side(f):
    def _(*args):
        check_lists = [a for a in args if hasattr(a, 'conn_side')]
        sides = [ConnSide.RT, ConnSide.BOTH]

        # TODO: Accurate Error Message
        assert all(a.conn_side in sides for a in check_lists)
        return f(*args)

    return _



import random
import uuid

from py_hcl.firrtl_ir.shortcuts import w, uw, u, s, sw, n, vec
from py_hcl.firrtl_ir.type import UIntType, SIntType, \
    UnknownType, BundleType, VectorType, ClockType
from py_hcl.firrtl_ir.type.field import Field
from py_hcl.firrtl_ir.type_checker import check
from py_hcl.utils import signed_num_bin_len, unsigned_num_bin_len


class OpCase(object):
    def __init__(self, op):
        self.op = op
        self.args = []
        self.const_args = []
        self.res_type_fn = None
        self.valid_filter = lambda *_: True

    def arg_types(self, *arg_types):
        self.args = arg_types
        return self

    def const_arg_types(self, *const_arg_types):
        self.const_args = const_arg_types
        return self

    def res_type(self, res_type_fn):
        self.res_type_fn = res_type_fn
        return self

    def filter(self, valid_fn):
        self.valid_filter = valid_fn
        return self


def int_gen():
    return random.randint(0, 20)


def name_gen():
    return "n" + str(uuid.uuid4()).replace("-", "")


def u_gen():
    if random.randint(0, 1):
        rand_u_value = random.randint(0, 1024)
        rand_u_value_width = unsigned_num_bin_len(rand_u_value)
        rand_u_width = random.randint(rand_u_value_width,
                                      2 * rand_u_value_width)
        return u(rand_u_value, w(rand_u_width))
    else:
        rand_nu_name = name_gen()
        rand_nu_width = random.randint(1, 20)
        return n(rand_nu_name, uw(rand_nu_width))


def s_gen():
    if random.randint(0, 1):
        rand_s_value = random.randint(-1024, 1024)
        rand_s_value_width = signed_num_bin_len(rand_s_value)
        rand_s_width = random.randint(rand_s_value_width,
                                      2 * rand_s_value_width)
        return s(rand_s_value, w(rand_s_width))
    else:
        rand_ns_name = name_gen()
        rand_ns_width = random.randint(1, 20)
        return n(rand_ns_name, sw(rand_ns_width))


def uw_gen():
    rand_width = random.randint(1, 20)
    return uw(rand_width)


def sw_gen():
    rand_width = random.randint(1, 20)
    return sw(rand_width)


def unknown_gen():
    return n(name_gen(), UnknownType())


def clock_gen():
    return n(name_gen(), ClockType())


def vec_gen():
    rand_num = random.randint(1, 100)
    seed = random.choices(population=[1, 2, 3, 4],
                          weights=[0.4, 0.4, 0.1, 0.1],
                          k=1)[0]
    if seed == 1:
        return vec(uw_gen(), rand_num)
    elif seed == 2:
        return vec(sw_gen(), rand_num)
    elif seed == 3:
        return vec(vec_gen(), rand_num)
    elif seed == 4:
        return vec(bdl_gen(), rand_num)


def bdl_gen():
    rand_num = random.randint(1, 5)
    fields = []
    for i in range(rand_num):
        rand_name = name_gen()
        rand_bool = random.choice([True, False])
        seed = random.choices(population=[1, 2, 3, 4],
                              weights=[0.4, 0.4, 0.1, 0.1],
                              k=1)[0]
        if seed == 1:
            fields.append(Field(rand_name, uw_gen(), rand_bool))
        elif seed == 2:
            fields.append(Field(rand_name, sw_gen(), rand_bool))
        elif seed == 3:
            fields.append(Field(rand_name, vec_gen(), rand_bool))
        elif seed == 4:
            fields.append(Field(rand_name, bdl_gen(), rand_bool))
    return BundleType(fields)


obj_gen_map = {
    UIntType: u_gen,
    SIntType: s_gen,
    UnknownType: unknown_gen,
    ClockType: clock_gen,
    VectorType: vec_gen,
    BundleType: bdl_gen,
    int: int_gen,
}


def obj_gen(case):
    while True:
        args = [obj_gen_map[a]() for a in case.args]
        const_args = [obj_gen_map[a]() for a in case.const_args]

        construct_args = []
        if len(args) > 1:
            construct_args.append(args)
        elif len(args) == 1:
            construct_args.append(args[0])

        if len(const_args) > 1:
            construct_args.append(const_args)
        elif len(const_args) == 1:
            construct_args.append(const_args[0])

        if case.valid_filter(*args, *const_args):
            break

    construct_args.append(case.res_type_fn(*args, *const_args))

    return case.op(*construct_args)


def basis_tester(cases):
    for case in cases:
        for i in range(5):
            obj = obj_gen(case)
            assert check(obj)


def encounter_error_tester(cases):
    for case in cases:
        for i in range(5):
            obj = obj_gen(case)
            assert not check(obj)


def op_args(op, *arg_types):
    class C:
        @staticmethod
        def tpe(res_type):
            return OpCase(op).arg_types(*arg_types).res_type(res_type)

    return C


def type_wrong_cases_2_args_gen(op):
    wrong_cases = [
        op_args(op, UnknownType, UnknownType).tpe(lambda x, y: uw(32)),
        op_args(op, UnknownType, UIntType).tpe(lambda x, y: uw(32)),
        op_args(op, UnknownType, SIntType).tpe(lambda x, y: sw(32)),
        op_args(op, UnknownType, VectorType).tpe(lambda x, y: sw(32)),
        op_args(op, UnknownType, BundleType).tpe(lambda x, y: sw(32)),
        op_args(op, UIntType, VectorType).tpe(lambda x, y: uw(32)),
        op_args(op, UIntType, BundleType).tpe(lambda x, y: uw(32)),
        op_args(op, UIntType, UnknownType).tpe(lambda x, y: uw(32)),
        op_args(op, SIntType, VectorType).tpe(lambda x, y: sw(32)),
        op_args(op, SIntType, BundleType).tpe(lambda x, y: sw(32)),
        op_args(op, SIntType, UnknownType).tpe(lambda x, y: sw(32)),
        op_args(op, VectorType, UIntType).tpe(lambda x, y: uw(32)),
        op_args(op, VectorType, SIntType).tpe(lambda x, y: sw(32)),
        op_args(op, VectorType, VectorType).tpe(lambda x, y: uw(32)),
        op_args(op, VectorType, BundleType).tpe(lambda x, y: uw(32)),
        op_args(op, BundleType, UIntType).tpe(lambda x, y: uw(32)),
        op_args(op, BundleType, SIntType).tpe(lambda x, y: uw(32)),
        op_args(op, BundleType, VectorType).tpe(lambda x, y: uw(32)),
        op_args(op, BundleType, UnknownType).tpe(lambda x, y: uw(32)),
        op_args(op, BundleType, BundleType).tpe(lambda x, y: uw(32)),
    ]
    return wrong_cases


def type_wrong_cases_1_arg_gen(op):
    wrong_cases = [
        op_args(op, UnknownType).tpe(lambda x: uw(32)),
        op_args(op, VectorType).tpe(lambda x: uw(32)),
        op_args(op, BundleType).tpe(lambda x: uw(32)),
    ]
    return wrong_cases


def width(x):
    return x.tpe.width.width


def max_width(x, y):
    return max(width(x), width(y))


def min_width(x, y):
    return min(width(x), width(y))


def sum_width(x, y):
    return width(x) + width(y)



from ...utils import serialize_str
from .. import Statement


class DefRegister(Statement):
    def __init__(self, name, tpe, clock_ref):
        self.name = name
        self.tpe = tpe
        self.clock_ref = clock_ref

    def serialize_stmt(self, output, indent):
        output.write(b"reg ")
        output.write(serialize_str(self.name))
        output.write(b" : ")
        self.tpe.serialize(output)
        output.write(b", ")
        self.clock_ref.serialize(output)


class DefInitRegister(Statement):
    def __init__(self, name, tpe, clock_ref, reset_ref, init_ref):
        self.name = name
        self.tpe = tpe
        self.clock_ref = clock_ref
        self.reset_ref = reset_ref
        self.init_ref = init_ref

    def serialize_stmt(self, output, indent):
        output.write(b"reg ")
        output.write(serialize_str(self.name))
        output.write(b" : ")
        self.tpe.serialize(output)
        output.write(b", ")
        self.clock_ref.serialize(output)
        output.write(b" with :\n")
        output.write(serialize_str("  " * (indent + 1)))
        output.write(b"reset => (")
        self.reset_ref.serialize(output)
        output.write(b", ")
        self.init_ref.serialize(output)
        output.write(b")")



from typing import Union

from py_hcl.core.type import HclType
from py_hcl.core.type.bundle import BundleT, Dir


def Bundle(**named_ports: Union[HclType, dict]) -> BundleT:
    t = {k: ({'dir': Dir.SRC, 'hcl_type': v} if isinstance(v, HclType) else v)
         for k, v in named_ports.items()}

    return BundleT(t)



from py_hcl.core.expr import HclExpr
from py_hcl.core.stmt.connect import ConnSide
from py_hcl.core.type.sint import SIntT
from py_hcl.utils import signed_num_bin_len


class SLiteral(HclExpr):
    def __init__(self, value: int):
        self.value = value

        w = signed_num_bin_len(value)
        self.hcl_type = SIntT(w)
        self.conn_side = ConnSide.RT



from io import BytesIO


def serialize_equal(serializable, target):
    output = BytesIO()
    serializable.serialize(output)
    output.flush()
    str_equal(str(output.getvalue(), "utf-8"), target)
    output.close()


def serialize_stmt_equal(serializable, target):
    output = BytesIO()
    serializable.serialize_stmt(output, 0)
    output.flush()
    str_equal(str(output.getvalue(), "utf-8"), target)
    output.close()


def str_equal(lf, rt):
    # for the more precise error report
    assert lf == rt



import logging

from multipledispatch import dispatch

from ..utils import type_in
from ...expr.literal import SIntLiteral, SIntType, UIntLiteral, UIntType
from py_hcl.utils import signed_num_bin_len, unsigned_num_bin_len

checker = dispatch


###############################################################
#                      TYPE CHECKERS                          #
###############################################################

@checker(UIntLiteral)
def check(uint: UIntLiteral):
    if not type_in(uint.tpe, UIntType):
        logging.error("uint: type check failed - {}".format(uint.tpe))
        return False

    if uint.value < 0:
        logging.error("uint: value check failed - {}".format(uint.value))
        return False

    if unsigned_num_bin_len(uint.value) > uint.tpe.width.width:
        logging.error("uint: width check failed - {} > {}".format(
            unsigned_num_bin_len(uint.value), uint.tpe.width.width)
        )
        return False

    return True


@checker(SIntLiteral)
def check(sint: SIntLiteral):
    if not type_in(sint.tpe, SIntType):
        logging.error("sint: type check failed - {}".format(sint.tpe))
        return False

    if signed_num_bin_len(sint.value) > sint.tpe.width.width:
        logging.error("sint: width check failed - {} > {}".format(
            signed_num_bin_len(sint.value), sint.tpe.width.width)
        )
        return False

    return True



from py_hcl.firrtl_ir.expr.literal import SIntLiteral, UIntLiteral
from py_hcl.firrtl_ir.type import UnknownType
from py_hcl.firrtl_ir.type_checker import check
from py_hcl.firrtl_ir.type.width import Width
from ..utils import serialize_equal


def test_uint_literal():
    ui = UIntLiteral(10, Width(4))
    assert check(ui)
    serialize_equal(ui, 'UInt<4>("ha")')

    ui = UIntLiteral(-10, Width(5))
    assert not check(ui)

    ui = UIntLiteral(1023, Width(10))
    assert check(ui)
    serialize_equal(ui, 'UInt<10>("h3ff")')

    ui = UIntLiteral(-1023, Width(11))
    assert not check(ui)

    ui = UIntLiteral(10, Width(3))
    assert not check(ui)

    ui = UIntLiteral(-10, Width(4))
    assert not check(ui)

    ui = UIntLiteral(10, Width(3))
    ui.tpe = UnknownType()
    assert not check(ui)


def test_sint_literal():
    si = SIntLiteral(10, Width(5))
    assert check(si)
    serialize_equal(si, 'SInt<5>("ha")')

    si = SIntLiteral(-10, Width(5))
    assert check(si)
    serialize_equal(si, 'SInt<5>("h-a")')

    si = SIntLiteral(1023, Width(11))
    assert check(si)
    serialize_equal(si, 'SInt<11>("h3ff")')

    si = SIntLiteral(-1023, Width(11))
    assert check(si)
    serialize_equal(si, 'SInt<11>("h-3ff")')

    si = SIntLiteral(10, Width(4))
    assert not check(si)

    si = SIntLiteral(-10, Width(4))
    assert not check(si)

    si = SIntLiteral(10, Width(5))
    si.tpe = UnknownType()
    assert not check(si)


def test_checker():
    assert not check(123)



from typing import Dict, Union, Optional, Tuple

from py_hcl.core.expr import HclExpr
from py_hcl.core.expr.error import ExprError
from py_hcl.core.stmt.connect import ConnSide
from py_hcl.core.type import HclType
from py_hcl.core.type.bundle import Dir, BundleT
from py_hcl.core.utils import module_inherit_mro
from py_hcl.utils import json_serialize


@json_serialize
class Input(object):
    def __init__(self, hcl_type: HclType):
        self.port_dir = 'input'
        self.hcl_type = hcl_type


@json_serialize
class Output(object):
    def __init__(self, hcl_type: HclType):
        self.port_dir = 'output'
        self.hcl_type = hcl_type


@json_serialize
class IOHolder(object):
    def __init__(self, named_ports: Dict[str, Union[Input, Output]],
                 module_name: Optional[str] = None):
        self.named_ports = named_ports
        self.module_name = module_name


@json_serialize
class IONode(object):
    def __init__(self, io_holder: IOHolder,
                 next_node: Optional["IOHolder"]):
        self.io_holder = io_holder
        if next_node is not None:
            self.next_node = next_node


@json_serialize(json_fields=["id", "type", "hcl_type",
                             "conn_side", "io_chain_head"])
class IO(HclExpr):
    def __init__(self, hcl_type: HclType, io_chain_head: IONode):
        self.type = 'io'
        self.hcl_type = hcl_type
        self.conn_side = ConnSide.RT
        self.io_chain_head = io_chain_head


def io_extend(modules: Tuple[type]):
    modules = module_inherit_mro(modules)

    current_ports = {}
    io_chain = None
    for m in modules[::-1]:
        h = m.io.io_chain_head.io_holder
        current_ports.update(h.named_ports)
        io_chain = IONode(h, io_chain)

    def _(named_ports: Dict[str, Union[Input, Output]]):
        current_ports.update(named_ports)
        io_chain_head = IONode(IOHolder(named_ports), io_chain)
        return IO(calc_type_from_ports(current_ports), io_chain_head)

    return _


def calc_type_from_ports(named_ports: Dict[str, Union[Input, Output]]):
    types = {}
    for k, v in named_ports.items():
        if isinstance(v, Input):
            types[k] = {"dir": Dir.SRC, "hcl_type": v.hcl_type}
            continue

        if isinstance(v, Output):
            types[k] = {"dir": Dir.SINK, "hcl_type": v.hcl_type}
            continue

        raise ExprError.io_value(
            "type of '{}' is {}, not Input or Output".format(k, type(v)))

    return BundleT(types)



from py_hcl.firrtl_ir.expr.prim_ops import Shr
from py_hcl.firrtl_ir.shortcuts import sw, uw, u, w, s
from py_hcl.firrtl_ir.type import BundleType, SIntType, \
    UIntType, UnknownType, VectorType
from tests.test_firrtl_ir.utils import serialize_equal
from .helper import basis_tester, encounter_error_tester, OpCase, width


def args(*arg_types):
    class A:
        @staticmethod
        def const(*const_types):
            class B:
                @staticmethod
                def tpe(res_type):
                    return OpCase(Shr) \
                        .arg_types(*arg_types) \
                        .const_arg_types(*const_types) \
                        .res_type(res_type)

            return B

    return A


shr_basis_cases = [
    args(UIntType).const(int).tpe(lambda x, y: uw(max(1, width(x) - y))),
    args(SIntType).const(int).tpe(lambda x, y: sw(max(1, width(x) - y))),
]

shr_type_wrong_cases = [
    args(UnknownType).const(int).tpe(lambda x, y: uw(max(1, y))),
    args(VectorType).const(int).tpe(lambda x, y: uw(max(1, y))),
    args(BundleType).const(int).tpe(lambda x, y: uw(max(1, y))),
]

shr_width_wrong_cases = [
    args(UIntType).const(int).tpe(lambda x, y: uw(max(1, width(x) - y) + 1)),
    args(SIntType).const(int).tpe(lambda x, y: sw(max(1, width(x) - y) + 1)),
    args(UIntType).const(int).tpe(lambda x, y: uw(max(1, width(x) - y) - 1)),
    args(SIntType).const(int).tpe(lambda x, y: sw(max(1, width(x) - y) - 1)),
]


def test_shr():
    basis_tester(shr_basis_cases)
    encounter_error_tester(shr_type_wrong_cases)
    encounter_error_tester(shr_width_wrong_cases)
    serialize_equal(Shr(u(20, w(5)), 3, uw(2)),
                    'shr(UInt<5>("h14"), 3)')
    serialize_equal(Shr(s(-20, w(6)), 3, uw(3)),
                    'shr(SInt<6>("h-14"), 3)')



from py_hcl.firrtl_ir.shortcuts import u, w, n, uw
from py_hcl.firrtl_ir.stmt.block import Block
from py_hcl.firrtl_ir.stmt.conditionally import Conditionally
from py_hcl.firrtl_ir.stmt.connect import Connect
from py_hcl.firrtl_ir.stmt.defn.node import DefNode
from py_hcl.firrtl_ir.stmt.empty import EmptyStmt
from py_hcl.firrtl_ir.type_checker import check
from ..utils import serialize_stmt_equal


def test_block_basis():
    blk = Block([EmptyStmt()])
    assert check(blk)
    serialize_stmt_equal(blk, "skip")

    blk = Block([DefNode("n", u(1, w(1))),
                 Conditionally(n("n", uw(1)),
                               EmptyStmt(),
                               Connect(n("a", uw(8)), n("b", uw(8))))
                 ])
    assert check(blk)
    serialize_stmt_equal(blk, 'node n = UInt<1>("h1")\n'
                              'when n :\n'
                              '  skip\n'
                              'else :\n'
                              '  a <= b')


def test_block_empty():
    blk = Block([])
    assert not check(blk)



from py_hcl.firrtl_ir.expr.accessor import SubField, SubIndex, SubAccess
from py_hcl.firrtl_ir.shortcuts import bdl, uw, sw, n, vec, u, w, s
from py_hcl.firrtl_ir.type_checker import check
from ..utils import serialize_equal


def test_sub_field_basis():
    bd = bdl(a=(uw(8), False), b=(sw(8), False))
    sf = SubField(n("bd", bd), "a", uw(8))
    assert check(sf)
    serialize_equal(sf, "bd.a")

    bd = bdl(a=(bdl(c=(uw(8), True)), False), b=(sw(8), False))
    sf = SubField(n("bd", bd), "a", bdl(c=(uw(8), True)))
    assert check(sf)
    serialize_equal(sf, "bd.a")


def test_sub_field_non_exist():
    bd = bdl(a=(uw(8), False), b=(sw(8), False))
    sf = SubField(n("bd", bd), "c", uw(8))
    assert not check(sf)

    bd = bdl(a=(bdl(c=(uw(8), True)), False), b=(sw(8), False))
    sf = SubField(n("bd", bd), "c", bdl(c=(uw(8), True)))
    assert not check(sf)


def test_sub_field_not_bundle():
    sf = SubField(n("bd", uw(8)), "c", uw(8))
    assert not check(sf)

    sf = SubField(n("bd", vec(uw(8), 32)), "c", bdl(c=(uw(8), True)))
    assert not check(sf)


def test_sub_field_type_wrong():
    bd = bdl(a=(uw(8), False), b=(sw(8), False))
    sf = SubField(n("bd", bd), "a", sw(8))
    assert not check(sf)

    bd = bdl(a=(bdl(c=(uw(8), True)), False), b=(sw(8), False))
    sf = SubField(n("bd", bd), "a", bdl(c=(uw(9), True)))
    assert not check(sf)


def test_sub_index_basis():
    vc = vec(uw(8), 10)
    si = SubIndex(n("vc", vc), 5, uw(8))
    assert check(si)
    serialize_equal(si, "vc[5]")

    vc = vec(vec(uw(8), 10), 20)
    si = SubIndex(n("vc", vc), 19, vec(uw(8), 10))
    assert check(si)
    serialize_equal(si, "vc[19]")


def test_sub_index_over_bound():
    vc = vec(uw(8), 10)
    si = SubIndex(n("vc", vc), 10, uw(8))
    assert not check(si)

    vc = vec(vec(uw(8), 10), 20)
    si = SubIndex(n("vc", vc), -1, vec(uw(8), 10))
    assert not check(si)


def test_sub_index_non_vector():
    si = SubIndex(n("vc", uw(8)), 5, uw(1))
    assert not check(si)

    si = SubIndex(n("vc", bdl(a=(vec(uw(8), 10), True))), 19, vec(uw(8), 10))
    assert not check(si)


def test_sub_index_type_wrong():
    vc = vec(uw(8), 10)
    si = SubIndex(n("vc", vc), 5, sw(8))
    assert not check(si)

    vc = vec(vec(uw(8), 10), 20)
    si = SubIndex(n("vc", vc), 19, vec(uw(8), 20))
    assert not check(si)


def test_sub_access_basis():
    vc = vec(uw(8), 10)
    sa = SubAccess(n("vc", vc), u(2, w(3)), uw(8))
    assert check(sa)
    serialize_equal(sa, 'vc[UInt<3>("h2")]')

    vc = vec(uw(8), 10)
    sa = SubAccess(n("vc", vc), n("a", uw(8)), uw(8))
    assert check(sa)
    serialize_equal(sa, 'vc[a]')


def test_sub_access_non_vector():
    sa = SubAccess(n("vc", uw(8)), u(2, w(3)), uw(8))
    assert not check(sa)

    sa = SubAccess(n("vc", bdl(a=(vec(uw(8), 10), True))),
                   u(2, w(3)), vec(uw(8), 10))
    assert not check(sa)


def test_sub_access_idx_non_uint():
    vc = vec(uw(8), 10)
    sa = SubAccess(n("vc", vc), s(2, w(3)), uw(8))
    assert not check(sa)

    vc = vec(uw(8), 10)
    sa = SubAccess(n("vc", vc), n("a", sw(8)), uw(8))
    assert not check(sa)



import pytest

from py_hcl.core.expr import ExprTable
from py_hcl.core.expr.error import ExprError
from py_hcl.core.type.bundle import BundleT
from py_hcl.dsl.expr.io import IO, Input, Output, io_extend
from py_hcl.dsl.module import Module
from py_hcl.core.type import HclType
from py_hcl.dsl.tpe.uint import U


class A(Module):
    io = IO(
        i=Input(U.w(8)),
        o=Output(U.w(8)))

    io.o <<= io.i


def test_io():
    table = A.packed_module.named_expr_chain.named_expr_chain_head \
        .named_expr_holder.named_expression_table
    id = list(table.keys())[list(table.values()).index('io')]
    t = ExprTable.table[id].hcl_type
    assert isinstance(t, BundleT)
    assert len(t.fields) == 2


def test_io_inherit_basis():
    class B(A):
        io = io_extend(A)(
            i1=Input(U.w(9)),
        )
        io.o <<= io.i1

    table = B.packed_module.named_expr_chain.named_expr_chain_head \
        .named_expr_holder.named_expression_table
    id = list(table.keys())[list(table.values()).index('io')]
    t = ExprTable.table[id].hcl_type
    assert isinstance(t, BundleT)
    assert len(t.fields) == 3


def test_io_inherit_override():
    class B(A):
        io = io_extend(A)(
            i=Input(U.w(9)),
        )
        io.o <<= io.i

    table = B.packed_module.named_expr_chain.named_expr_chain_head \
        .named_expr_holder.named_expression_table
    id = list(table.keys())[list(table.values()).index('io')]
    t = ExprTable.table[id].hcl_type
    assert isinstance(t, BundleT)
    assert len(t.fields) == 2


def test_io_no_wrap_io():
    with pytest.raises(ExprError, match='^.*Input.*Output.*$'):
        class A(Module):
            io = IO(i=HclType())

    with pytest.raises(ExprError, match='^.*Input.*Output.*$'):
        class A(Module):
            io = IO(
                i=HclType(),
                o=Output(HclType()))

    with pytest.raises(ExprError, match='^.*Input.*Output.*$'):
        class A(Module):
            io = IO(
                i=Input(HclType()),
                o=HclType())



from py_hcl import *  # noqa: F403
from py_hcl.transformer.pyhcl_to_firrtl.convertor import convert


class A(Module):
    io = IO(
        sel=Input(U.w(2)),
        i1=Input(U.w(3)),
        i2=Input(U.w(3)),
        o=Output(S.w(8)),
    )

    io.o <<= U(8)
    with when(io.sel[0]):
        io.o <<= io.i1
    with else_when(io.sel[1]):
        with when(io.i1[2]):
            io.o <<= io.i1 + io.sel
    with otherwise():
        io.o <<= io.i1 + io.i2


class C(Module):
    io = IO(
        i3=Input(U.w(20)),
        o3=Output(U.w(10))
    )

    io.o3 <<= io.i3


class B(A):
    io = io_extend(A)(
        i3=Input(U.w(20)),
        o3=Output(U.w(10))
    )

    c = C()
    c.i3 <<= io.i3
    io.o3 <<= c.o3


class FullAdder(Module):
    io = IO(
        a=Input(Bool),
        b=Input(Bool),
        cin=Input(Bool),
        sum=Output(Bool),
        cout=Output(Bool),
    )

    # Generate the sum
    io.sum <<= io.a ^ io.b ^ io.cin

    # Generate the carry
    io.cout <<= io.a & io.b | io.b & io.cin | io.a & io.cin


if __name__ == '__main__':
    compile_to_firrtl(FullAdder)



from ..utils import serialize_str


class Field(object):
    def __init__(self, name, tpe, is_flipped=False):
        self.name = name
        self.tpe = tpe
        self.is_flipped = is_flipped

    def serialize(self, output):
        if self.is_flipped:
            output.write(b"flip ")
        output.write(serialize_str(self.name))
        output.write(b" : ")
        self.tpe.serialize(output)



from enum import Enum

from py_hcl.utils import json_serialize


class ScopeType(Enum):
    TOP = 0
    GROUND = 1
    WHEN = 2
    ELSE_WHEN = 3
    OTHERWISE = 4


@json_serialize
class ScopeInfo(object):
    def __init__(self, scope_id, scope_level, scope_type, tag_object=None):
        self.scope_id = scope_id
        self.scope_level = scope_level
        self.scope_type = scope_type
        if tag_object:
            self.tag_object = tag_object


class ScopeLevelManager(object):
    _next_scope_level = 0

    @classmethod
    def current_level(cls):
        return cls._next_scope_level

    @classmethod
    def expand_level(cls):
        cls._next_scope_level += 1

    @classmethod
    def shrink_level(cls):
        cls._next_scope_level -= 1


class ScopeIdManager(object):
    _next_scope_id = 0

    @classmethod
    def next_id(cls):
        ret = cls._next_scope_id
        cls._next_scope_id += 1
        return ret


class ScopeManager(object):
    scope_list = [
        ScopeInfo(
            # TOP SCOPE which contains all modules
            scope_id=ScopeIdManager.next_id(),
            scope_level=ScopeLevelManager.current_level(),
            scope_type=ScopeType.TOP,
        )
    ]
    scope_id_map = {
        scope_list[0].scope_id: scope_list[0]
    }
    scope_expanding_hooks = []
    scope_shrinking_hooks = []

    @classmethod
    def expand_scope(cls, scope_type, tag_object=None):
        ScopeLevelManager.expand_level()

        current_scope = cls.current_scope()
        next_id = ScopeIdManager.next_id()
        next_scope = ScopeInfo(
            scope_id=next_id,
            scope_level=ScopeLevelManager.current_level(),
            scope_type=scope_type,
            tag_object=tag_object,
        )
        cls.scope_id_map[next_id] = next_scope
        cls.scope_list.append(next_scope)

        for fn in cls.scope_expanding_hooks:
            fn(current_scope, next_scope)

    @classmethod
    def shrink_scope(cls):
        ScopeLevelManager.shrink_level()
        current_scope = cls.scope_list.pop()
        next_scope = cls.current_scope()

        for fn in cls.scope_shrinking_hooks:
            fn(current_scope, next_scope)

    @classmethod
    def current_scope(cls):
        return cls.scope_list[-1]

    @classmethod
    def get_scope_info(cls, sid):
        return cls.scope_id_map[sid]

    @classmethod
    def register_scope_expanding(cls, fn):
        cls.scope_expanding_hooks.append(fn)

    @classmethod
    def register_scope_shrinking(cls, fn):
        cls.scope_shrinking_hooks.append(fn)



from py_hcl.firrtl_ir.shortcuts import uw, sw, vec, bdl
from py_hcl.firrtl_ir.type import UnknownType, ClockType
from py_hcl.firrtl_ir.type_measurer import equal


def test_type_eq():
    assert equal(UnknownType(), UnknownType())
    assert equal(ClockType(), ClockType())
    assert equal(uw(10), uw(10))
    assert equal(sw(10), sw(10))
    assert equal(vec(uw(10), 8), vec(uw(10), 8))
    assert equal(bdl(a=(vec(uw(10), 8), False), b=(uw(10), True)),
                 bdl(a=(vec(uw(10), 8), False), b=(uw(10), True)))


def test_type_neq():
    assert not equal(UnknownType(), ClockType())
    assert not equal(UnknownType(), uw(10))
    assert not equal(UnknownType(), sw(10))
    assert not equal(UnknownType(), vec(uw(10), 8))
    assert not equal(UnknownType(), vec(sw(10), 8))
    assert not equal(UnknownType(),
                     bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)))

    assert not equal(ClockType(), UnknownType())
    assert not equal(ClockType(), uw(10))
    assert not equal(ClockType(), sw(10))
    assert not equal(ClockType(), vec(uw(10), 8))
    assert not equal(ClockType(), vec(sw(10), 8))
    assert not equal(ClockType(),
                     bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)))

    assert not equal(uw(10), UnknownType())
    assert not equal(uw(10), ClockType())
    assert not equal(uw(10), uw(8))
    assert not equal(uw(10), sw(10))
    assert not equal(uw(10), sw(10))
    assert not equal(uw(10), vec(uw(10), 8))
    assert not equal(uw(10), vec(sw(10), 8))
    assert not equal(uw(10),
                     bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)))

    assert not equal(sw(10), UnknownType())
    assert not equal(sw(10), ClockType())
    assert not equal(sw(10), sw(8))
    assert not equal(sw(10), uw(10))
    assert not equal(sw(10), uw(10))
    assert not equal(sw(10), vec(uw(10), 8))
    assert not equal(sw(10), vec(sw(10), 8))
    assert not equal(sw(10),
                     bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)))

    assert not equal(vec(uw(10), 8), UnknownType())
    assert not equal(vec(uw(10), 8), ClockType())
    assert not equal(vec(uw(10), 8), sw(8))
    assert not equal(vec(uw(10), 8), uw(10))
    assert not equal(vec(uw(10), 8), uw(10))
    assert not equal(vec(uw(10), 8), vec(uw(8), 8))
    assert not equal(vec(uw(10), 8), vec(uw(10), 9))
    assert not equal(vec(uw(10), 8), vec(sw(10), 8))
    assert not equal(vec(uw(10), 8),
                     bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)))

    assert not equal(bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)),
                     UnknownType())
    assert not equal(bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)),
                     ClockType())
    assert not equal(bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)),
                     sw(8))
    assert not equal(bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)),
                     uw(10))
    assert not equal(bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)),
                     uw(10))
    assert not equal(bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)),
                     vec(sw(10), 8))
    assert not equal(bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)),
                     bdl(a=(vec(uw(10), 8), True), b=(uw(10), False)))
    assert not equal(bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)),
                     bdl(a=(vec(uw(10), 2), False), b=(uw(10), False)))
    assert not equal(bdl(a=(uw(3), False), b=(uw(10), False)),
                     bdl(a=(uw(3), False), b=(uw(10), False), c=(sw(2), True)))
    assert not equal(bdl(a=(vec(uw(10), 8), False), b=(uw(10), False)),
                     bdl(b=(uw(10), False), a=(vec(uw(10), 8), False)))



from py_hcl.utils import json_serialize


@json_serialize(json_fields=['stmt_class', 'statement'])
class LineStatement(object):
    def __init__(self, scope_id, statement):
        self.stmt_class = 'line'
        self.scope_id = scope_id
        self.statement = statement


@json_serialize(json_fields=['stmt_class', 'statements'])
class ClusterStatement(object):
    def __init__(self, scope_info, stmts):
        self.stmt_class = 'cluster'
        self.scope_info = scope_info
        self.statements = stmts


@json_serialize
class ConditionStatement(object):
    def __init__(self, seq_cond_id, seq_stmts, alt_stmts):
        self.stmt_class = 'condition'
        self.seq_cond_id = seq_cond_id
        self.seq_stmts = seq_stmts
        self.alt_stmts = alt_stmts



from py_hcl.core.expr import HclExpr
from py_hcl.core.stmt.connect import ConnSide
from py_hcl.utils import json_serialize


@json_serialize(
    json_fields=['id', 'type', 'hcl_type', "conn_side", 'module_name'])
class ModuleInst(HclExpr):
    def __init__(self, module_cls):
        self.type = 'module_inst'
        self.hcl_type = module_cls.io.hcl_type
        self.conn_side = ConnSide.LF
        self.packed_module = module_cls.packed_module
        self.module_name = module_cls.packed_module.name


def con_module(module_cls):
    return ModuleInst(module_cls)



from py_hcl.core.error import CoreError


def set_up():
    StatementError.append({
        'WrongBranchSyntax': {
            'code': 300,
            'value': StatementError(
                'expected a well-defined when-else_when-otherwise block')},
        'ConnectTypeError': {
            'code': 301,
            'value': StatementError(
                'connect statement contains unexpected types')
        }
    })


class StatementError(CoreError):
    @staticmethod
    def wrong_branch_syntax(msg):
        return StatementError.err('WrongBranchSyntax', msg)

    @staticmethod
    def connect_type_error(*args):
        ts = ', '.join([type(a.hcl_type).__name__ for a in args])
        msg = 'connect(): unsupported connect types: {}'.format(ts)
        return StatementError.err('ConnectTypeError', msg)


set_up()



from .expr.literal import UIntLiteral, SIntLiteral
from .expr.reference import Reference
from .type import SIntType, UIntType, VectorType, BundleType
from .type.field import Field
from .type.width import Width


def sw(width):
    return SIntType(Width(width))


def uw(width):
    return UIntType(Width(width))


def w(width):
    return Width(width)


def u(value, width):
    return UIntLiteral(value, width)


def s(value, width):
    return SIntLiteral(value, width)


def n(name, tpe):
    return Reference(name, tpe)


def vec(tpe, size):
    return VectorType(tpe, size)


def bdl(**field):
    fields = [Field(k, field[k][0], field[k][1]) for k in field]
    return BundleType(fields)



import logging

from multipledispatch import dispatch

from ...expr.accessor import SubField, SubIndex, SubAccess
from ...type import BundleType, VectorType, UIntType
from ...type_checker.utils import type_in
from ...type_measurer import equal

checker = dispatch


###############################################################
#                      TYPE CHECKERS                          #
###############################################################

@checker(SubField)
def check(sub_field):
    from .. import check_all_expr
    if not check_all_expr(sub_field.bundle_ref):
        logging.error("sub_field: reference check failed - {}".format(
            sub_field.bundle_ref
        ))
        return False

    if not type_in(sub_field.bundle_ref.tpe, BundleType):
        logging.error("sub_field: reference type check failed - {}".format(
            sub_field.bundle_ref
        ))
        return False

    field = None
    for f in sub_field.bundle_ref.tpe.fields:
        if f.name == sub_field.name:
            field = f
            break
    if field is None:
        logging.error("sub_field: field not exist - {} not in {}".format(
            sub_field.name, sub_field.bundle_ref.tpe.fields
        ))
        return False

    if not equal(field.tpe, sub_field.tpe):
        return False

    return True


@checker(SubIndex)
def check(sub_index):
    from .. import check_all_expr
    if not check_all_expr(sub_index.vector_ref):
        return False

    if not type_in(sub_index.vector_ref.tpe, VectorType):
        return False

    if sub_index.vector_ref.tpe.size <= sub_index.index or sub_index.index < 0:
        return False

    if not equal(sub_index.vector_ref.tpe.elem_type, sub_index.tpe):
        return False

    return True


@checker(SubAccess)
def check(sub_access):
    from .. import check_all_expr
    if not check_all_expr(sub_access.vector_ref, sub_access.index_ref):

        return False

    if not type_in(sub_access.vector_ref.tpe, VectorType):
        return False

    if not type_in(sub_access.index_ref.tpe, UIntType):
        return False

    if not equal(sub_access.vector_ref.tpe.elem_type, sub_access.tpe):
        return False

    return True



from ..utils import serialize_str, serialize_num
from . import Expression


class SubField(Expression):
    def __init__(self, bundle_ref, name, tpe):
        self.bundle_ref = bundle_ref
        self.name = name
        self.tpe = tpe

    def serialize(self, output):
        self.bundle_ref.serialize(output)
        output.write(b".")
        output.write(serialize_str(self.name))


class SubIndex(Expression):
    def __init__(self, vector_ref, index, tpe):
        self.vector_ref = vector_ref
        self.index = index
        self.tpe = tpe

    def serialize(self, output):
        self.vector_ref.serialize(output)
        output.write(b"[")
        output.write(serialize_num(self.index))
        output.write(b"]")


class SubAccess(Expression):
    def __init__(self, vector_ref, index_ref, tpe):
        self.vector_ref = vector_ref
        self.index_ref = index_ref
        self.tpe = tpe

    def serialize(self, output):
        self.vector_ref.serialize(output)
        output.write(b"[")
        self.index_ref.serialize(output)
        output.write(b"]")



from multipledispatch.dispatcher import MethodDispatcher

from py_hcl.core.hcl_ops import op_apply
from py_hcl.core.stmt.connect import ConnSide
from py_hcl.core.type import UnknownType, HclType
from py_hcl.utils import json_serialize


class ExprIdGen:
    id = -1

    @classmethod
    def next_id(cls):
        cls.id += 1
        return cls.id


class ExprTable:
    table = {}

    @classmethod
    def add(cls, i, obj):
        cls.table[i] = obj

    @classmethod
    def get(cls, i):
        return cls.table[i]


@json_serialize
class HclExpr(object):
    hcl_type = UnknownType()
    conn_side = ConnSide.UNKNOWN

    def __new__(cls, *args):
        obj = super().__new__(cls)
        obj.id = ExprIdGen.next_id()
        ExprTable.add(obj.id, obj)
        return obj

    def __ilshift__(self, other):
        return op_apply('<<=')(self, other)

    def __add__(self, other):
        return op_apply('+')(self, other)

    def __and__(self, other):
        return op_apply('&')(self, other)

    def __or__(self, other):
        return op_apply('|')(self, other)

    def __xor__(self, other):
        return op_apply('^')(self, other)

    def __getattr__(self, item):
        return op_apply('.')(self, item)

    def __setitem__(self, key, value):
        return self[key]

    __getitem__ = MethodDispatcher('__getitem__')

    @__getitem__.register(tuple)
    def _(self, item):
        if len(item) == 1:
            return self[item[0]]
        return self.__getitem__(item[0])[item[1:]]

    @__getitem__.register(slice)
    def _(self, item):
        start = item.start if item.start is not None else 0
        if item.step is None:
            if item.stop is not None:
                return op_apply('[i:j]')(self, start, item.stop)
            if item.stop is None:
                return op_apply('[i:]')(self, start)
        if item.stop is not None:
            return op_apply('[i:j:k]')(self, start, item.stop, item.step)
        if item.stop is None:
            return op_apply('[i::k]')(self, start, item.step)

    @__getitem__.register(int)
    def _(self, item):
        return op_apply('[i]')(self, item)

    def to_uint(self):
        return op_apply('to_uint')(self)

    def to_sint(self):
        return op_apply('to_sint')(self)

    def to_bool(self):
        return op_apply('to_bool')(self)


@json_serialize(json_fields=['id', 'type', 'hcl_type', 'conn_side', 'op_node'])
class ExprHolder(HclExpr):
    def __init__(self, hcl_type: HclType, conn_side: ConnSide, op_node):
        self.type = 'expr_holder'
        self.hcl_type = hcl_type
        self.conn_side = conn_side
        self.op_node = op_node



from multipledispatch import dispatch

from py_hcl.core.type import UnknownType as HclUnknownType
from py_hcl.core.type.bundle import BundleT, Dir
from py_hcl.core.type.clock import ClockT
from py_hcl.core.type.sint import SIntT
from py_hcl.core.type.uint import UIntT
from py_hcl.core.type.vector import VectorT
from py_hcl.firrtl_ir.type import SIntType, UIntType, UnknownType, ClockType, \
    VectorType, BundleType
from py_hcl.firrtl_ir.type.field import Field
from py_hcl.firrtl_ir.type.width import Width


@dispatch()
def convert_type(_: HclUnknownType):
    return UnknownType()


@dispatch()
def convert_type(sint: SIntT):
    return SIntType(Width(sint.width))


@dispatch()
def convert_type(uint: UIntT):
    return UIntType(Width(uint.width))


@dispatch()
def convert_type(_: ClockT):
    return ClockType()


@dispatch()
def convert_type(vec: VectorT):
    return VectorType(convert_type(vec.inner_type), vec.size)


@dispatch()
def convert_type(bundle: BundleT):
    fields = []
    for k, v in bundle.fields.items():
        f = Field(k, convert_type(v['hcl_type']), v['dir'] == Dir.SINK)
        fields.append(f)
    return BundleType(fields)



from setuptools import setup, find_packages

__version__ = "0.1.2"
install_requires = ['multipledispatch']
tests_require = ['pytest', 'pytest-cov']
setup_requires = ['pytest-runner', 'multipledispatch']

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="py_hcl",
    version=__version__,
    author="scutdig",
    author_email="zhongzc_arch@outlook.com",
    description=("A Hardware Construct Language"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/scutdig/pyhcl",
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=setup_requires,
    packages=find_packages(),
    python_requires='>=3.6',
)



from ...utils import serialize_str
from .. import Statement


class DefNode(Statement):
    def __init__(self, name, expr_ref):
        self.name = name
        self.expr_ref = expr_ref

    def serialize_stmt(self, output, indent):
        output.write(b"node ")
        output.write(serialize_str(self.name))
        output.write(b" = ")
        self.expr_ref.serialize(output)



import py_hcl.core.stmt.branch as cb
from py_hcl.core.expr import HclExpr


class when(object):
    def __init__(self, cond_expr: HclExpr):
        self.cond_expr = cond_expr

    def __enter__(self):
        cb.do_when_enter(self.cond_expr)

    def __exit__(self, exc_type, exc_val, exc_tb):
        cb.do_when_exit()


class else_when(object):
    def __init__(self, cond_expr: HclExpr):
        self.cond_expr = cond_expr

    def __enter__(self):
        cb.do_else_when_enter(self.cond_expr)

    def __exit__(self, exc_type, exc_val, exc_tb):
        cb.do_else_when_exit()


class otherwise(object):
    def __enter__(self):
        cb.do_otherwise_enter()

    def __exit__(self, exc_type, exc_val, exc_tb):
        cb.do_otherwise_exit()
